# ruff: noqa: F821

import sys
import warnings
import textwrap
from enum import Enum
from typing import Any, Optional, Union
from typing_extensions import Self

import pytest

import strawberry

pytestmark = pytest.mark.skipif(
    sys.version_info < (3, 12), reason="These are tests for Python 3.12+"
)


def test_supports_generic_simple_type():
    @strawberry.type
    class Edge[T]:
        cursor: strawberry.ID
        node_field: T

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> Edge[int]:
            return Edge(cursor=strawberry.ID("1"), node_field=1)

    schema = strawberry.Schema(query=Query)

    query = """{
        example {
            __typename
            cursor
            nodeField
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "example": {"__typename": "IntEdge", "cursor": "1", "nodeField": 1}
    }


def test_supports_generic_specialized():
    @strawberry.type
    class Edge[T]:
        cursor: strawberry.ID
        node_field: T

    @strawberry.type
    class IntEdge(Edge[int]): ...

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> IntEdge:
            return IntEdge(cursor=strawberry.ID("1"), node_field=1)

    schema = strawberry.Schema(query=Query)

    query = """{
        example {
            __typename
            cursor
            nodeField
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "example": {"__typename": "IntEdge", "cursor": "1", "nodeField": 1}
    }


def test_supports_generic_specialized_subclass():
    @strawberry.type
    class Edge[T]:
        cursor: strawberry.ID
        node_field: T

    @strawberry.type
    class IntEdge(Edge[int]): ...

    @strawberry.type
    class IntEdgeSubclass(IntEdge): ...

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> IntEdgeSubclass:
            return IntEdgeSubclass(cursor=strawberry.ID("1"), node_field=1)

    schema = strawberry.Schema(query=Query)

    query = """{
        example {
            __typename
            cursor
            nodeField
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "example": {"__typename": "IntEdgeSubclass", "cursor": "1", "nodeField": 1}
    }


def test_supports_generic_specialized_with_type():
    @strawberry.type
    class Fruit:
        name: str

    @strawberry.type
    class Edge[T]:
        cursor: strawberry.ID
        node_field: T

    @strawberry.type
    class FruitEdge(Edge[Fruit]): ...

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> FruitEdge:
            return FruitEdge(cursor=strawberry.ID("1"), node_field=Fruit(name="Banana"))

    schema = strawberry.Schema(query=Query)

    query = """{
        example {
            __typename
            cursor
            nodeField {
                name
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "example": {
            "__typename": "FruitEdge",
            "cursor": "1",
            "nodeField": {"name": "Banana"},
        }
    }


def test_supports_generic_specialized_with_list_type():
    @strawberry.type
    class Fruit:
        name: str

    @strawberry.type
    class Edge[T]:
        cursor: strawberry.ID
        nodes: list[T]

    @strawberry.type
    class FruitEdge(Edge[Fruit]): ...

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> FruitEdge:
            return FruitEdge(
                cursor=strawberry.ID("1"),
                nodes=[Fruit(name="Banana"), Fruit(name="Apple")],
            )

    schema = strawberry.Schema(query=Query)

    query = """{
        example {
            __typename
            cursor
            nodes {
                name
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "example": {
            "__typename": "FruitEdge",
            "cursor": "1",
            "nodes": [
                {"name": "Banana"},
                {"name": "Apple"},
            ],
        }
    }


def test_supports_generic():
    @strawberry.type
    class Edge[T]:
        cursor: strawberry.ID
        node: T

    @strawberry.type
    class Person:
        name: str

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> Edge[Person]:
            return Edge(cursor=strawberry.ID("1"), node=Person(name="Example"))

    schema = strawberry.Schema(query=Query)

    query = """{
        example {
            __typename
            cursor
            node {
                name
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "example": {
            "__typename": "PersonEdge",
            "cursor": "1",
            "node": {"name": "Example"},
        }
    }


def test_supports_multiple_generic():
    @strawberry.type
    class Multiple[A, B]:
        a: A
        b: B

    @strawberry.type
    class Query:
        @strawberry.field
        def multiple(self) -> Multiple[int, str]:
            return Multiple(a=123, b="123")

    schema = strawberry.Schema(query=Query)

    query = """{
        multiple {
            __typename
            a
            b
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "multiple": {"__typename": "IntStrMultiple", "a": 123, "b": "123"}
    }


def test_support_nested_generics():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Edge[T]:
        node: T

    @strawberry.type
    class Connection[T]:
        edge: Edge[T]

    @strawberry.type
    class Query:
        @strawberry.field
        def users(self) -> Connection[User]:
            return Connection(edge=Edge(node=User(name="Patrick")))

    schema = strawberry.Schema(query=Query)

    query = """{
        users {
            __typename
            edge {
                __typename
                node {
                    name
                }
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "users": {
            "__typename": "UserConnection",
            "edge": {"__typename": "UserEdge", "node": {"name": "Patrick"}},
        }
    }


def test_supports_optional():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Edge[T]:
        node: Optional[T] = None

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> Edge[User]:
            return Edge()

    schema = strawberry.Schema(query=Query)

    query = """{
        user {
            __typename
            node {
                name
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"user": {"__typename": "UserEdge", "node": None}}


def test_supports_lists():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Edge[T]:
        nodes: list[T]

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> Edge[User]:
            return Edge(nodes=[])

    schema = strawberry.Schema(query=Query)

    query = """{
        user {
            __typename
            nodes {
                name
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"user": {"__typename": "UserEdge", "nodes": []}}


def test_supports_lists_of_optionals():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Edge[T]:
        nodes: list[Optional[T]]

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> Edge[User]:
            return Edge(nodes=[None])

    schema = strawberry.Schema(query=Query)

    query = """{
        user {
            __typename
            nodes {
                name
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"user": {"__typename": "UserEdge", "nodes": [None]}}


def test_can_extend_generics():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Edge[T]:
        node: T

    @strawberry.type
    class Connection[T]:
        edges: list[Edge[T]]

    @strawberry.type
    class ConnectionWithMeta[T](Connection[T]):
        meta: str

    @strawberry.type
    class Query:
        @strawberry.field
        def users(self) -> ConnectionWithMeta[User]:
            return ConnectionWithMeta(
                meta="123", edges=[Edge(node=User(name="Patrick"))]
            )

    schema = strawberry.Schema(query=Query)

    query = """{
        users {
            __typename
            meta
            edges {
                __typename
                node {
                    name
                }
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "users": {
            "__typename": "UserConnectionWithMeta",
            "meta": "123",
            "edges": [{"__typename": "UserEdge", "node": {"name": "Patrick"}}],
        }
    }


def test_supports_generic_in_unions():
    @strawberry.type
    class Edge[T]:
        cursor: strawberry.ID
        node: T

    @strawberry.type
    class Fallback:
        node: str

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> Union[Fallback, Edge[int]]:
            return Edge(cursor=strawberry.ID("1"), node=1)

    schema = strawberry.Schema(query=Query)

    query = """{
        example {
            __typename

            ... on IntEdge {
                cursor
                node
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "example": {"__typename": "IntEdge", "cursor": "1", "node": 1}
    }


def test_generic_with_enum_as_param_of_type_inside_unions():
    @strawberry.type
    class Pet:
        name: str

    @strawberry.type
    class ErrorNode[T]:
        code: T

    @strawberry.enum
    class Codes(Enum):
        a = "a"
        b = "b"

    @strawberry.type
    class Query:
        @strawberry.field
        def result(self) -> Union[Pet, ErrorNode[Codes]]:
            return ErrorNode(code=Codes.a)

    schema = strawberry.Schema(query=Query)

    query = """{
        result {
            __typename
            ... on CodesErrorNode {
                code
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"result": {"__typename": "CodesErrorNode", "code": "a"}}


def test_generic_with_enum():
    @strawberry.enum
    class EstimatedValueEnum(Enum):
        test = "test"
        testtest = "testtest"

    @strawberry.type
    class EstimatedValue[T]:
        value: T
        type: EstimatedValueEnum

    @strawberry.type
    class Query:
        @strawberry.field
        def estimated_value(self) -> Optional[EstimatedValue[int]]:
            return EstimatedValue(value=1, type=EstimatedValueEnum.test)

    schema = strawberry.Schema(query=Query)

    query = """{
        estimatedValue {
            __typename
            value
            type
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "estimatedValue": {
            "__typename": "IntEstimatedValue",
            "value": 1,
            "type": "test",
        }
    }


def test_supports_generic_in_unions_multiple_vars():
    @strawberry.type
    class Edge[A, B]:
        info: A
        node: B

    @strawberry.type
    class Fallback:
        node: str

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> Union[Fallback, Edge[int, str]]:
            return Edge(node="string", info=1)

    schema = strawberry.Schema(query=Query)

    query = """{
        example {
            __typename

            ... on IntStrEdge {
                node
                info
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "example": {"__typename": "IntStrEdge", "node": "string", "info": 1}
    }


def test_supports_generic_in_unions_with_nesting():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Edge[T]:
        node: T

    @strawberry.type
    class Connection[T]:
        edge: Edge[T]

    @strawberry.type
    class Fallback:
        node: str

    @strawberry.type
    class Query:
        @strawberry.field
        def users(self) -> Union[Connection[User], Fallback]:
            return Connection(edge=Edge(node=User(name="Patrick")))

    schema = strawberry.Schema(query=Query)

    query = """{
        users {
            __typename
            ... on UserConnection {
                edge {
                    __typename
                    node {
                        name
                    }
                }
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "users": {
            "__typename": "UserConnection",
            "edge": {"__typename": "UserEdge", "node": {"name": "Patrick"}},
        }
    }


def test_supports_multiple_generics_in_union():
    @strawberry.type
    class Edge[T]:
        cursor: strawberry.ID
        node: T

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> list[Union[Edge[int], Edge[str]]]:
            return [
                Edge(cursor=strawberry.ID("1"), node=1),
                Edge(cursor=strawberry.ID("2"), node="string"),
            ]

    schema = strawberry.Schema(query=Query)

    expected_schema = """
      type IntEdge {
        cursor: ID!
        node: Int!
      }

      union IntEdgeStrEdge = IntEdge | StrEdge

      type Query {
        example: [IntEdgeStrEdge!]!
      }

      type StrEdge {
        cursor: ID!
        node: String!
      }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()

    query = """{
        example {
            __typename

            ... on IntEdge {
                cursor
                intNode: node
            }

            ... on StrEdge {
                cursor
                strNode: node
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "example": [
            {"__typename": "IntEdge", "cursor": "1", "intNode": 1},
            {"__typename": "StrEdge", "cursor": "2", "strNode": "string"},
        ]
    }


def test_generated_names():
    @strawberry.type
    class EdgeWithCursor[T]:
        cursor: strawberry.ID
        node: T

    @strawberry.type
    class SpecialPerson:
        name: str

    @strawberry.type
    class Query:
        @strawberry.field
        def person_edge(self) -> EdgeWithCursor[SpecialPerson]:
            return EdgeWithCursor(
                cursor=strawberry.ID("1"), node=SpecialPerson(name="Example")
            )

    schema = strawberry.Schema(query=Query)

    query = """{
        personEdge {
            __typename
            cursor
            node {
                name
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "personEdge": {
            "__typename": "SpecialPersonEdgeWithCursor",
            "cursor": "1",
            "node": {"name": "Example"},
        }
    }


def test_supports_lists_within_unions():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Edge[T]:
        nodes: list[T]

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> Union[User, Edge[User]]:
            return Edge(nodes=[User(name="P")])

    schema = strawberry.Schema(query=Query)

    query = """{
        user {
            __typename

            ... on UserEdge {
                nodes {
                    name
                }
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"user": {"__typename": "UserEdge", "nodes": [{"name": "P"}]}}


def test_supports_lists_within_unions_empty_list():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Edge[T]:
        nodes: list[T]

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> Union[User, Edge[User]]:
            return Edge(nodes=[])

    schema = strawberry.Schema(query=Query)

    query = """{
        user {
            __typename

            ... on UserEdge {
                nodes {
                    name
                }
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"user": {"__typename": "UserEdge", "nodes": []}}


@pytest.mark.xfail
def test_raises_error_when_unable_to_find_type():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Edge[T]:
        nodes: list[T]

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> Union[User, Edge[User]]:
            return Edge(nodes=["bad example"])  # type: ignore

    schema = strawberry.Schema(query=Query)

    query = """{
        user {
            __typename

            ... on UserEdge {
                nodes {
                    name
                }
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert result.errors[0].message == (
        "Unable to find type for <class 'tests.schema.test_generics."
        "test_raises_error_when_unable_to_find_type.<locals>.Edge'> "
        "and (<class 'str'>,)"
    )


def test_generic_with_arguments():
    @strawberry.type
    class Collection[T]:
        @strawberry.field
        def by_id(self, ids: list[int]) -> list[T]:
            return []

    @strawberry.type
    class Post:
        name: str

    @strawberry.type
    class Query:
        user: Collection[Post]

    schema = strawberry.Schema(Query)

    expected_schema = """
    type Post {
      name: String!
    }

    type PostCollection {
      byId(ids: [Int!]!): [Post!]!
    }

    type Query {
      user: PostCollection!
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()


def test_generic_argument():
    @strawberry.type
    class Node[T]:
        @strawberry.field
        def edge(self, arg: T) -> bool:
            return bool(arg)

        @strawberry.field
        def edges(self, args: list[T]) -> int:
            return len(args)

    @strawberry.type
    class Query:
        i_node: Node[int]
        b_node: Node[bool]

    schema = strawberry.Schema(Query)

    expected_schema = """
    type BoolNode {
      edge(arg: Boolean!): Boolean!
      edges(args: [Boolean!]!): Int!
    }

    type IntNode {
      edge(arg: Int!): Boolean!
      edges(args: [Int!]!): Int!
    }

    type Query {
      iNode: IntNode!
      bNode: BoolNode!
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()


def test_generic_extra_type():
    @strawberry.type
    class Node[T]:
        field: T

    @strawberry.type
    class Query:
        name: str

    schema = strawberry.Schema(Query, types=[Node[int]])

    expected_schema = """
    type IntNode {
      field: Int!
    }

    type Query {
      name: String!
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()


def test_generic_extending_with_type_var():
    @strawberry.interface
    class Node[T]:
        id: strawberry.ID

        def _resolve(self) -> Optional[T]:
            return None

    @strawberry.type
    class Book(Node[str]):
        name: str

    @strawberry.type
    class Query:
        @strawberry.field
        def books(self) -> list[Book]:
            return []

    schema = strawberry.Schema(query=Query)

    expected_schema = """
    type Book implements Node {
      id: ID!
      name: String!
    }

    interface Node {
      id: ID!
    }

    type Query {
      books: [Book!]!
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()


def test_self():
    @strawberry.interface
    class INode:
        field: Optional[Self]
        fields: list[Self]

    @strawberry.type
    class Node(INode): ...

    schema = strawberry.Schema(query=Node)

    expected_schema = """
    schema {
      query: Node
    }

    interface INode {
      field: INode
      fields: [INode!]!
    }

    type Node implements INode {
      field: Node
      fields: [Node!]!
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()

    query = """{
        field {
            __typename
        }
        fields {
            __typename
        }
    }"""
    result = schema.execute_sync(query, root_value=Node(field=None, fields=[]))
    assert result.data == {"field": None, "fields": []}


def test_supports_generic_input_type():
    @strawberry.input
    class Input[T]:
        field: T

    @strawberry.type
    class Query:
        @strawberry.field
        def field(self, input: Input[str]) -> str:
            return input.field

    schema = strawberry.Schema(query=Query)

    query = """{
        field(input: { field: "data" })
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"field": "data"}


def test_generic_interface():
    @strawberry.interface
    class ObjectType:
        obj: strawberry.Private[Any]

        @strawberry.field
        def repr(self) -> str:
            return str(self.obj)

    @strawberry.type
    class GenericObject[T](ObjectType):
        @strawberry.field
        def value(self) -> T:
            return self.obj

    @strawberry.type
    class Query:
        @strawberry.field
        def foo(self) -> GenericObject[str]:
            return GenericObject(obj="foo")

    schema = strawberry.Schema(query=Query)
    query_result = schema.execute_sync(
        """
            query {
                foo {
                    __typename
                    value
                    repr
                }
            }
        """
    )

    assert not query_result.errors
    assert query_result.data == {
        "foo": {
            "__typename": "StrGenericObject",
            "value": "foo",
            "repr": "foo",
        }
    }


def test_generic_interface_extra_types():
    @strawberry.interface
    class Abstract:
        x: str = ""

    @strawberry.type
    class Real[T](Abstract):
        y: T

    @strawberry.type
    class Query:
        @strawberry.field
        def real(self) -> Abstract:
            return Real[int](y=0)

    schema = strawberry.Schema(Query, types=[Real[int]])

    assert (
        str(schema)
        == textwrap.dedent(
            """
            interface Abstract {
              x: String!
            }

            type IntReal implements Abstract {
              x: String!
              y: Int!
            }

            type Query {
              real: Abstract!
            }
            """
        ).strip()
    )

    query_result = schema.execute_sync("{ real { __typename x } }")

    assert not query_result.errors
    assert query_result.data == {"real": {"__typename": "IntReal", "x": ""}}


def test_generics_via_anonymous_union():
    @strawberry.type
    class Edge[T]:
        cursor: str
        node: T

    @strawberry.type
    class Connection[T]:
        edges: list[Edge[T]]

    @strawberry.type
    class Entity1:
        id: int

    @strawberry.type
    class Entity2:
        id: int

    @strawberry.type
    class Query:
        entities: Connection[Union[Entity1, Entity2]]

    schema = strawberry.Schema(query=Query)

    expected_schema = textwrap.dedent(
        """
        type Entity1 {
          id: Int!
        }

        union Entity1Entity2 = Entity1 | Entity2

        type Entity1Entity2Connection {
          edges: [Entity1Entity2Edge!]!
        }

        type Entity1Entity2Edge {
          cursor: String!
          node: Entity1Entity2!
        }

        type Entity2 {
          id: Int!
        }

        type Query {
          entities: Entity1Entity2Connection!
        }
        """
    ).strip()

    assert str(schema) == expected_schema


def test_info_type_alias_syntax():
    type MyInfo = strawberry.Info[None, None]  # type: ignore[valid-type]

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self, info: MyInfo) -> str:
            return "Hello World"

    # Should not raise DeprecationWarning about name-based matching
    with warnings.catch_warnings():
        warnings.filterwarnings("error", category=DeprecationWarning)
        schema = strawberry.Schema(query=Query)

    result = schema.execute_sync("query { hello }")
    assert result.errors is None
    assert result.data == {"hello": "Hello World"}



from textwrap import dedent

import pytest

import strawberry
from strawberry.tools import merge_types


@strawberry.type
class Person:
    @strawberry.field
    def name(self) -> str:
        return "Eve"

    @strawberry.field
    def age(self) -> int:
        return 42


@strawberry.type
class SimpleGreeter:
    @strawberry.field
    def hi(self) -> str:
        return "Hi"


@strawberry.type
class ComplexGreeter:
    @strawberry.field
    def hi(self, name: str = "world") -> str:
        return f"Hello, {name}!"

    @strawberry.field
    def bye(self, name: str = "world") -> str:
        return f"Bye, {name}!"


def test_custom_name():
    """The resulting type should have a custom name is one is specified"""
    custom_name = "SuperQuery"
    ComboQuery = merge_types(custom_name, (ComplexGreeter, Person))
    assert ComboQuery.__name__ == custom_name


def test_inheritance():
    """It should merge multiple types following the regular inheritance rules"""
    ComboQuery = merge_types("SuperType", (ComplexGreeter, Person))

    definition = ComboQuery.__strawberry_definition__
    assert len(definition.fields) == 4

    actuals = [(field.python_name, field.type) for field in definition.fields]
    expected = [("hi", str), ("bye", str), ("name", str), ("age", int)]
    assert actuals == expected


def test_empty_list():
    """It should raise when the `types` argument is empty"""
    with pytest.raises(ValueError):
        merge_types("EmptyType", ())


def test_schema():
    """It should create a valid, usable schema based on a merged query"""
    ComboQuery = merge_types("SuperSchema", (ComplexGreeter, Person))
    schema = strawberry.Schema(query=ComboQuery)

    sdl = """
        schema {
          query: SuperSchema
        }

        type SuperSchema {
          hi(name: String! = "world"): String!
          bye(name: String! = "world"): String!
          name: String!
          age: Int!
        }
    """
    assert dedent(sdl).strip() == str(schema)

    result = schema.execute_sync("query { hi }")
    assert not result.errors
    assert result.data == {"hi": "Hello, world!"}


def test_fields_override():
    """It should warn when merging results in overriding fields"""
    with pytest.warns(Warning):  # noqa: PT030
        merge_types("FieldsOverride", (ComplexGreeter, SimpleGreeter))



from inline_snapshot import snapshot

from .utils.marks import requires_mypy, requires_pyright, requires_ty, skip_on_windows
from .utils.typecheck import Result, typecheck

pytestmark = [skip_on_windows, requires_pyright, requires_mypy, requires_ty]


CODE = """
import strawberry


@strawberry.type
class SomeType:
    foobar: strawberry.auto


obj1 = SomeType(foobar=1)
obj2 = SomeType(foobar="some text")
obj3 = SomeType(foobar={"some key": "some value"})

reveal_type(obj1.foobar)
reveal_type(obj2.foobar)
reveal_type(obj3.foobar)
"""


def test_auto():
    result = typecheck(CODE)

    assert result.pyright == snapshot(
        [
            Result(
                type="information",
                message='Type of "obj1.foobar" is "Any"',
                line=14,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "obj2.foobar" is "Any"',
                line=15,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "obj3.foobar" is "Any"',
                line=16,
                column=13,
            ),
        ]
    )

    assert result.mypy == snapshot(
        [
            Result(type="note", message='Revealed type is "Any"', line=14, column=13),
            Result(type="note", message='Revealed type is "Any"', line=15, column=13),
            Result(type="note", message='Revealed type is "Any"', line=16, column=13),
        ]
    )
    assert result.ty == snapshot(
        [
            Result(
                type="information",
                message="Revealed type: `Any`",
                line=14,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `Any`",
                line=15,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `Any`",
                line=16,
                column=13,
            ),
        ]
    )



import dataclasses
import textwrap
import warnings
from collections.abc import Iterable
from typing_extensions import Self

import pytest
from pytest_mock import MockerFixture

import strawberry
from strawberry import relay
from strawberry.annotation import StrawberryAnnotation
from strawberry.relay.fields import ConnectionExtension
from strawberry.relay.utils import to_base64
from strawberry.types.arguments import StrawberryArgument
from strawberry.types.field import StrawberryField
from strawberry.types.fields.resolver import StrawberryResolver

from .schema import FruitAsync, schema


def test_query_node():
    result = schema.execute_sync(
        """
        query TestQuery ($id: ID!) {
            node (id: $id) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "id": to_base64("Fruit", 2),
        },
    )
    assert result.errors is None
    assert result.data == {
        "node": {
            "id": to_base64("Fruit", 2),
            "color": "red",
            "name": "Apple",
        },
    }


async def test_query_node_with_async_permissions():
    result = await schema.execute(
        """
        query TestQuery ($id: ID!) {
            nodeWithAsyncPermissions (id: $id) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "id": to_base64("Fruit", 2),
        },
    )
    assert result.errors is None
    assert result.data == {
        "nodeWithAsyncPermissions": {
            "id": to_base64("Fruit", 2),
            "color": "red",
            "name": "Apple",
        },
    }


def test_query_node_optional():
    result = schema.execute_sync(
        """
        query TestQuery ($id: ID!) {
            nodeOptional (id: $id) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "id": to_base64("Fruit", 999),
        },
    )
    assert result.errors is None
    assert result.data == {"nodeOptional": None}


async def test_query_node_async():
    result = await schema.execute(
        """
        query TestQuery ($id: ID!) {
            node (id: $id) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "id": to_base64("Fruit", 2),
        },
    )
    assert result.errors is None
    assert result.data == {
        "node": {
            "id": to_base64("Fruit", 2),
            "color": "red",
            "name": "Apple",
        },
    }


async def test_query_node_optional_async():
    result = await schema.execute(
        """
        query TestQuery ($id: ID!) {
            nodeOptional (id: $id) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "id": to_base64("Fruit", 999),
        },
    )
    assert result.errors is None
    assert result.data == {"nodeOptional": None}


def test_query_nodes():
    result = schema.execute_sync(
        """
        query TestQuery ($ids: [ID!]!) {
            nodes (ids: $ids) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "ids": [to_base64("Fruit", 2), to_base64("Fruit", 4)],
        },
    )
    assert result.errors is None
    assert result.data == {
        "nodes": [
            {
                "id": to_base64("Fruit", 2),
                "name": "Apple",
                "color": "red",
            },
            {
                "id": to_base64("Fruit", 4),
                "name": "Grape",
                "color": "purple",
            },
        ],
    }


def test_query_nodes_optional():
    result = schema.execute_sync(
        """
        query TestQuery ($ids: [ID!]!) {
            nodesOptional (ids: $ids) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "ids": [
                to_base64("Fruit", 2),
                to_base64("Fruit", 999),
                to_base64("Fruit", 4),
            ],
        },
    )
    assert result.errors is None
    assert result.data == {
        "nodesOptional": [
            {
                "id": to_base64("Fruit", 2),
                "name": "Apple",
                "color": "red",
            },
            None,
            {
                "id": to_base64("Fruit", 4),
                "name": "Grape",
                "color": "purple",
            },
        ],
    }


async def test_query_nodes_async():
    result = await schema.execute(
        """
        query TestQuery ($ids: [ID!]!) {
            nodes (ids: $ids) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
                ... on FruitAsync {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "ids": [
                to_base64("Fruit", 2),
                to_base64("Fruit", 4),
                to_base64("FruitAsync", 2),
            ],
        },
    )
    assert result.errors is None
    assert result.data == {
        "nodes": [
            {
                "id": to_base64("Fruit", 2),
                "name": "Apple",
                "color": "red",
            },
            {
                "id": to_base64("Fruit", 4),
                "name": "Grape",
                "color": "purple",
            },
            {
                "id": to_base64("FruitAsync", 2),
                "name": "Apple",
                "color": "red",
            },
        ],
    }


async def test_query_nodes_optional_async():
    result = await schema.execute(
        """
        query TestQuery ($ids: [ID!]!) {
            nodesOptional (ids: $ids) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
                ... on FruitAsync {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "ids": [
                to_base64("Fruit", 2),
                to_base64("FruitAsync", 999),
                to_base64("Fruit", 4),
                to_base64("Fruit", 999),
                to_base64("FruitAsync", 2),
            ],
        },
    )
    assert result.errors is None
    assert result.data == {
        "nodesOptional": [
            {
                "id": to_base64("Fruit", 2),
                "name": "Apple",
                "color": "red",
            },
            None,
            {
                "id": to_base64("Fruit", 4),
                "name": "Grape",
                "color": "purple",
            },
            None,
            {
                "id": to_base64("FruitAsync", 2),
                "name": "Apple",
                "color": "red",
            },
        ],
    }


fruits_query = """
query TestQuery (
    $first: Int = null
    $last: Int = null
    $before: String = null,
    $after: String = null,
) {{
    {} (
        first: $first
        last: $last
        before: $before
        after: $after
    ) {{
        pageInfo {{
            hasNextPage
            hasPreviousPage
            startCursor
            endCursor
        }}
        edges {{
            cursor
            node {{
                id
                name
                color
            }}
        }}
    }}
}}
"""

attrs = [
    "fruits",
    "fruitsLazy",
    "fruitsAlias",
    "fruitsAliasLazy",
    "fruitsConcreteResolver",
    "fruitsCustomResolver",
    "fruitsCustomResolverLazy",
    "fruitsCustomResolverIterator",
    "fruitsCustomResolverIterable",
    "fruitsCustomResolverGenerator",
    "fruitAlikeConnectionCustomResolver",
]
async_attrs = [
    *attrs,
    "fruitsAsync",
    "fruitsCustomResolverAsyncIterator",
    "fruitsCustomResolverAsyncIterable",
    "fruitsCustomResolverAsyncGenerator",
]


@pytest.mark.parametrize("query_attr", attrs)
def test_query_connection(query_attr: str):
    result = schema.execute_sync(
        fruits_query.format(query_attr),
        variable_values={},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 1),
                        "color": "yellow",
                        "name": "Banana",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjQ=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "4"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", async_attrs)
async def test_query_connection_async(mocker, query_attr: str):
    mocker.patch.object(FruitAsync, "resolve_typename", return_value="Fruit")

    result = await schema.execute(
        fruits_query.format(query_attr),
        variable_values={},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 1),
                        "color": "yellow",
                        "name": "Banana",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjQ=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "4"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", attrs)
def test_query_connection_filtering_first(query_attr: str):
    result = schema.execute_sync(
        fruits_query.format(query_attr),
        variable_values={"first": 2},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 1),
                        "color": "yellow",
                        "name": "Banana",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "1"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", async_attrs)
async def test_query_connection_filtering_first_async(mocker, query_attr: str):
    mocker.patch.object(FruitAsync, "resolve_typename", return_value="Fruit")

    result = await schema.execute(
        fruits_query.format(query_attr),
        variable_values={"first": 2},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 1),
                        "color": "yellow",
                        "name": "Banana",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "1"),
            },
        }
    }


def test_query_connection_filtering_after_without_first():
    result = schema.execute_sync(
        """{ someFruits {
            edges { node { id } }
            pageInfo {
                endCursor
            }
        } }"""
    )
    assert not result.errors
    assert len(result.data["someFruits"]["edges"]) == 100
    assert (
        relay.from_base64(result.data["someFruits"]["edges"][99]["node"]["id"])[1]
        == "99"
    )
    result = schema.execute_sync(
        """query ($after: String!){ someFruits(after: $after, first: 100) {
            edges { node { id } }
        } }""",
        variable_values={"after": result.data["someFruits"]["pageInfo"]["endCursor"]},
    )
    assert not result.errors
    assert len(result.data["someFruits"]["edges"]) == 100
    assert (
        relay.from_base64(result.data["someFruits"]["edges"][-1]["node"]["id"])[1]
        == "199"
    )


@pytest.mark.parametrize("query_attr", attrs)
def test_query_connection_filtering_first_with_after(query_attr: str):
    result = schema.execute_sync(
        fruits_query.format(query_attr),
        variable_values={"first": 2, "after": to_base64("arrayconnection", "1")},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", async_attrs)
async def test_query_connection_filtering_first_with_after_async(
    mocker, query_attr: str
):
    mocker.patch.object(FruitAsync, "resolve_typename", return_value="Fruit")

    result = await schema.execute(
        fruits_query.format(query_attr),
        variable_values={"first": 2, "after": to_base64("arrayconnection", "1")},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", attrs)
def test_query_connection_filtering_last(query_attr: str):
    result = schema.execute_sync(
        fruits_query.format(query_attr),
        variable_values={"last": 2},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjQ=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "3"),
                "endCursor": to_base64("arrayconnection", "4"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", async_attrs)
async def test_query_connection_filtering_last_async(mocker, query_attr: str):
    mocker.patch.object(FruitAsync, "resolve_typename", return_value="Fruit")

    result = await schema.execute(
        fruits_query.format(query_attr),
        variable_values={"last": 2},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjQ=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "3"),
                "endCursor": to_base64("arrayconnection", "4"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", attrs)
def test_query_connection_filtering_first_with_before(query_attr: str):
    result = schema.execute_sync(
        fruits_query.format(query_attr),
        variable_values={"first": 1, "before": to_base64("arrayconnection", "3")},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "2"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", async_attrs)
async def test_query_connection_filtering_first_with_before_async(
    mocker, query_attr: str
):
    mocker.patch.object(FruitAsync, "resolve_typename", return_value="Fruit")

    result = await schema.execute(
        fruits_query.format(query_attr),
        variable_values={"first": 1, "before": to_base64("arrayconnection", "3")},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "2"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", attrs)
def test_query_connection_filtering_last_with_before(query_attr: str):
    result = schema.execute_sync(
        fruits_query.format(query_attr),
        variable_values={"last": 2, "before": to_base64("arrayconnection", "4")},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", async_attrs)
async def test_query_connection_filtering_last_with_before_async(
    mocker, query_attr: str
):
    mocker.patch.object(FruitAsync, "resolve_typename", return_value="Fruit")

    result = await schema.execute(
        fruits_query.format(query_attr),
        variable_values={"last": 2, "before": to_base64("arrayconnection", "4")},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        }
    }


fruits_custom_query = """
query TestQuery (
    $first: Int = null
    $last: Int = null
    $before: String = null,
    $after: String = null,
) {
    fruitsCustomPagination (
        first: $first
        last: $last
        before: $before
        after: $after
    ) {
        something
        pageInfo {
            hasNextPage
            hasPreviousPage
            startCursor
            endCursor
        }
        edges {
            cursor
            node {
                id
                name
                color
            }
        }
    }
}
"""


def test_query_custom_connection():
    result = schema.execute_sync(
        fruits_custom_query,
        variable_values={},
    )
    assert result.errors is None
    assert result.data == {
        "fruitsCustomPagination": {
            "something": "foobar",
            "edges": [
                {
                    "cursor": "ZnJ1aXRfbmFtZTpBcHBsZQ==",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
                {
                    "cursor": "ZnJ1aXRfbmFtZTpCYW5hbmE=",
                    "node": {
                        "id": to_base64("Fruit", 1),
                        "color": "yellow",
                        "name": "Banana",
                    },
                },
                {
                    "cursor": "ZnJ1aXRfbmFtZTpHcmFwZQ==",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "ZnJ1aXRfbmFtZTpPcmFuZ2U=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
                {
                    "cursor": "ZnJ1aXRfbmFtZTpQaW5lYXBwbGU=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
            ],
            "pageInfo": {
                "startCursor": to_base64("fruit_name", "Apple"),
                "endCursor": to_base64("fruit_name", "Pineapple"),
                "hasNextPage": False,
                "hasPreviousPage": False,
            },
        }
    }


def test_query_custom_connection_filtering_first():
    result = schema.execute_sync(
        fruits_custom_query,
        variable_values={"first": 2},
    )
    assert result.errors is None
    assert result.data == {
        "fruitsCustomPagination": {
            "something": "foobar",
            "edges": [
                {
                    "cursor": "ZnJ1aXRfbmFtZTpBcHBsZQ==",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
                {
                    "cursor": "ZnJ1aXRfbmFtZTpCYW5hbmE=",
                    "node": {
                        "id": to_base64("Fruit", 1),
                        "color": "yellow",
                        "name": "Banana",
                    },
                },
            ],
            "pageInfo": {
                "startCursor": to_base64("fruit_name", "Apple"),
                "endCursor": to_base64("fruit_name", "Banana"),
                "hasNextPage": True,
                "hasPreviousPage": False,
            },
        }
    }


def test_query_custom_connection_filtering_first_with_after():
    result = schema.execute_sync(
        fruits_custom_query,
        variable_values={"first": 2, "after": to_base64("fruit_name", "Banana")},
    )
    assert result.errors is None
    assert result.data == {
        "fruitsCustomPagination": {
            "something": "foobar",
            "edges": [
                {
                    "cursor": "ZnJ1aXRfbmFtZTpHcmFwZQ==",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "ZnJ1aXRfbmFtZTpPcmFuZ2U=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("fruit_name", "Grape"),
                "endCursor": to_base64("fruit_name", "Orange"),
            },
        }
    }


def test_query_custom_connection_filtering_last():
    result = schema.execute_sync(
        fruits_custom_query,
        variable_values={"last": 2},
    )
    assert result.errors is None
    assert result.data == {
        "fruitsCustomPagination": {
            "something": "foobar",
            "edges": [
                {
                    "cursor": "ZnJ1aXRfbmFtZTpPcmFuZ2U=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
                {
                    "cursor": "ZnJ1aXRfbmFtZTpQaW5lYXBwbGU=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": True,
                "startCursor": to_base64("fruit_name", "Orange"),
                "endCursor": to_base64("fruit_name", "Pineapple"),
            },
        }
    }


def test_query_custom_connection_filtering_last_with_before():
    result = schema.execute_sync(
        fruits_custom_query,
        variable_values={
            "last": 2,
            "before": to_base64("fruit_name", "Pineapple"),
        },
    )
    assert result.errors is None
    assert result.data == {
        "fruitsCustomPagination": {
            "something": "foobar",
            "edges": [
                {
                    "cursor": "ZnJ1aXRfbmFtZTpHcmFwZQ==",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "ZnJ1aXRfbmFtZTpPcmFuZ2U=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("fruit_name", "Grape"),
                "endCursor": to_base64("fruit_name", "Orange"),
            },
        }
    }


fruits_query_custom_resolver = """
query TestQuery (
    $first: Int = null
    $last: Int = null
    $before: String = null,
    $after: String = null,
    $nameEndswith: String = null
) {{
    {} (
        first: $first
        last: $last
        before: $before
        after: $after
        nameEndswith: $nameEndswith
    ) {{
        pageInfo {{
            hasNextPage
            hasPreviousPage
            startCursor
            endCursor
        }}
        edges {{
            cursor
            node {{
                id
                name
                color
            }}
        }}
    }}
}}
"""

custom_attrs = [
    "fruitsConcreteResolver",
    "fruitsCustomResolver",
    "fruitsCustomResolverIterator",
    "fruitsCustomResolverIterable",
    "fruitsCustomResolverGenerator",
    "fruitAlikeConnectionCustomResolver",
]
custom_async_attrs = [
    *attrs,
    "fruitsCustomResolverAsyncIterator",
    "fruitsCustomResolverAsyncIterable",
    "fruitsCustomResolverAsyncGenerator",
]


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_connection_custom_resolver(query_attr: str):
    result = schema.execute_sync(
        fruits_query_custom_resolver.format(query_attr),
        variable_values={"nameEndswith": "e"},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_connection_custom_resolver_filtering_first(query_attr: str):
    result = schema.execute_sync(
        fruits_query_custom_resolver.format(query_attr),
        variable_values={"first": 2, "nameEndswith": "e"},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "1"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_connection_custom_resolver_filtering_first_with_after(query_attr: str):
    result = schema.execute_sync(
        fruits_query_custom_resolver.format(query_attr),
        variable_values={
            "first": 2,
            "after": to_base64("arrayconnection", "1"),
            "nameEndswith": "e",
        },
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_connection_custom_resolver_filtering_last(query_attr: str):
    result = schema.execute_sync(
        fruits_query_custom_resolver.format(query_attr),
        variable_values={"last": 2, "nameEndswith": "e"},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_connection_custom_resolver_filtering_last_with_before(query_attr: str):
    result = schema.execute_sync(
        fruits_query_custom_resolver.format(query_attr),
        variable_values={
            "last": 2,
            "before": to_base64("arrayconnection", "3"),
            "nameEndswith": "e",
        },
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "1"),
                "endCursor": to_base64("arrayconnection", "2"),
            },
        }
    }


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_first_negative(query_attr: str):
    result = schema.execute_sync(
        fruits_query_custom_resolver.format(query_attr),
        variable_values={"first": -1},
    )
    assert result.errors is not None
    assert (
        result.errors[0].message == "Argument 'first' must be a non-negative integer."
    )


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_first_higher_than_max_results(query_attr: str):
    result = schema.execute_sync(
        fruits_query_custom_resolver.format(query_attr),
        variable_values={"first": 500},
    )
    assert result.errors is not None
    assert result.errors[0].message == "Argument 'first' cannot be higher than 100."


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_last_negative(query_attr: str):
    result = schema.execute_sync(
        fruits_query_custom_resolver.format(query_attr),
        variable_values={"last": -1},
    )
    assert result.errors is not None
    assert result.errors[0].message == "Argument 'last' must be a non-negative integer."


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_last_higher_than_max_results(query_attr: str):
    result = schema.execute_sync(
        fruits_query_custom_resolver.format(query_attr),
        variable_values={"last": 500},
    )
    assert result.errors is not None
    assert result.errors[0].message == "Argument 'last' cannot be higher than 100."


def test_parameters(mocker: MockerFixture):
    class CustomField(StrawberryField):
        @property
        def arguments(self) -> list[StrawberryArgument]:
            return [
                *super().arguments,
                StrawberryArgument(
                    python_name="foo",
                    graphql_name=None,
                    type_annotation=StrawberryAnnotation(str),
                    default=None,
                ),
            ]

        @arguments.setter
        def arguments(self, value: list[StrawberryArgument]):
            cls = self.__class__
            return super(cls, cls).arguments.fset(self, value)

    @strawberry.type
    class Fruit(relay.Node):
        code: relay.NodeID[str]

    def resolver(info: strawberry.Info) -> list[Fruit]: ...

    @strawberry.type
    class Query:
        fruit: relay.ListConnection[Fruit] = relay.connection(
            resolver=resolver,
            extensions=[ConnectionExtension()],
        )
        fruit_custom_field: relay.ListConnection[Fruit] = CustomField(
            base_resolver=StrawberryResolver(resolver),
            extensions=[ConnectionExtension()],
        )

    schema = strawberry.Schema(query=Query)
    expected = '''
    type Fruit implements Node {
      """The Globally Unique ID of this object"""
      id: ID!
    }

    """A connection to a list of items."""
    type FruitConnection {
      """Pagination data for this connection"""
      pageInfo: PageInfo!

      """Contains the nodes in this connection"""
      edges: [FruitEdge!]!
    }

    """An edge in a connection."""
    type FruitEdge {
      """A cursor for use in pagination"""
      cursor: String!

      """The item at the end of the edge"""
      node: Fruit!
    }

    """An object with a Globally Unique ID"""
    interface Node {
      """The Globally Unique ID of this object"""
      id: ID!
    }

    """Information to aid in pagination."""
    type PageInfo {
      """When paginating forwards, are there more items?"""
      hasNextPage: Boolean!

      """When paginating backwards, are there more items?"""
      hasPreviousPage: Boolean!

      """When paginating backwards, the cursor to continue."""
      startCursor: String

      """When paginating forwards, the cursor to continue."""
      endCursor: String
    }

    type Query {
      fruit(
        """Returns the items in the list that come before the specified cursor."""
        before: String = null

        """Returns the items in the list that come after the specified cursor."""
        after: String = null

        """Returns the first n items from the list."""
        first: Int = null

        """Returns the items in the list that come after the specified cursor."""
        last: Int = null
      ): FruitConnection!
      fruitCustomField(
        foo: String!

        """Returns the items in the list that come before the specified cursor."""
        before: String = null

        """Returns the items in the list that come after the specified cursor."""
        after: String = null

        """Returns the first n items from the list."""
        first: Int = null

        """Returns the items in the list that come after the specified cursor."""
        last: Int = null
      ): FruitConnection!
    }
    '''
    assert str(schema) == textwrap.dedent(expected).strip()


before_after_test_query = """
query fruitsBeforeAfterTest (
    $before: String = null,
    $after: String = null,
) {
    fruits (
        before: $before
        after: $after
    ) {
        edges {
            cursor
            node {
                id
            }
        }
    }
}
"""


async def test_query_before_error():
    """Verify if the error raised on a non-existing before hash
    raises the correct error.
    """
    # with pytest.raises(ValueError):
    index = to_base64("Fake", 9292292)
    result = await schema.execute(
        before_after_test_query,
        variable_values={"before": index},
    )
    assert result.errors is not None
    assert "Argument 'before' contains a non-existing value" in str(result.errors)


def test_query_after_error():
    """Verify if the error raised on a non-existing before hash
    raises the correct error.
    """
    index = to_base64("Fake", 9292292)
    result = schema.execute_sync(
        before_after_test_query,
        variable_values={"after": index},
    )

    assert result.errors is not None
    assert "Argument 'after' contains a non-existing value" in str(result.errors)


@pytest.mark.parametrize(
    ("type_name", "should_have_name"),
    [("Fruit", False), ("PublicFruit", True)],
)
@pytest.mark.django_db(transaction=True)
def test_correct_model_returned(type_name: str, should_have_name: bool):
    @dataclasses.dataclass
    class FruitModel:
        id: str
        name: str

    fruits: dict[str, FruitModel] = {"1": FruitModel(id="1", name="Strawberry")}

    @strawberry.type
    class Fruit(relay.Node):
        id: relay.NodeID[int]

        @classmethod
        def resolve_nodes(
            cls,
            *,
            info: strawberry.Info | None = None,
            node_ids: Iterable[str],
            required: bool = False,
        ) -> Iterable[Self | FruitModel | None]:
            return [fruits[nid] if required else fruits.get(nid) for nid in node_ids]

    @strawberry.type
    class PublicFruit(relay.Node):
        id: relay.NodeID[int]
        name: str

        @classmethod
        def resolve_nodes(
            cls,
            *,
            info: strawberry.Info | None = None,
            node_ids: Iterable[str],
            required: bool = False,
        ) -> Iterable[Self | FruitModel | None]:
            return [fruits[nid] if required else fruits.get(nid) for nid in node_ids]

    @strawberry.type
    class Query:
        node: relay.Node = relay.node()

    schema = strawberry.Schema(query=Query, types=[Fruit, PublicFruit])

    node_id = relay.to_base64(type_name, "1")
    result = schema.execute_sync(
        """
        query NodeQuery($id: ID!) {
          node(id: $id) {
            __typename
            id
            ... on PublicFruit {
              name
            }
          }
        }
    """,
        {"id": node_id},
    )
    assert result.errors is None
    assert isinstance(result.data, dict)

    assert result.data["node"]["__typename"] == type_name
    assert result.data["node"]["id"] == node_id
    if should_have_name:
        assert result.data["node"]["name"] == "Strawberry"
    else:
        assert "name" not in result.data["node"]


def test_relay_node_no_deprecation_warning():
    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "Hello World"

        node: relay.Node = relay.node()

    # Should not raise DeprecationWarning about name-based matching
    with warnings.catch_warnings():
        warnings.filterwarnings("error", category=DeprecationWarning)
        schema = strawberry.Schema(query=Query)

    result = schema.execute_sync(
        """
        query { hello }
        """
    )
    assert result.errors is None
    assert result.data == {"hello": "Hello World"}


def test_relay_node_on_nested_type():
    """Test that relay.node() fields on nested types can be instantiated without args.

    Regression test for https://github.com/strawberry-graphql/strawberry/issues/3129
    """
    from .schema import Fruit

    @strawberry.type
    class NestedType:
        node: relay.Node = relay.node()
        nodes: list[relay.Node] = relay.node()

    @strawberry.type
    class Query:
        @strawberry.field
        def nested(self) -> NestedType:
            # This should work without passing node/nodes args
            return NestedType()

    schema = strawberry.Schema(query=Query, types=[Fruit])

    result = schema.execute_sync(
        """
        query {
            nested {
                __typename
            }
        }
        """
    )
    assert result.errors is None
    assert result.data == {"nested": {"__typename": "NestedType"}}



import contextlib
import json
import uuid
from io import BytesIO

import pytest

from tests.views.schema import schema

from .clients.base import HttpClient


def _encode_multipart_formdata(fields, boundary=None):
    # Do we really need this?
    # Copied from urllib3: https://github.com/urllib3/urllib3/blob/main/src/urllib3/filepost.py
    if boundary is None:
        boundary = uuid.uuid4().hex

    body = []

    for fieldname, value in fields.items():
        body.append(f"--{boundary}\r\n".encode())

        if isinstance(value, tuple):
            filename, data, content_type = value
            body.append(
                f'Content-Disposition: form-data; name="{fieldname}"; '
                f'filename="{filename}"\r\n'.encode()
            )
            body.append(f"Content-Type: {content_type}\r\n\r\n".encode())

            if isinstance(data, str):
                data = data.encode("utf-8")
            body.append(data)
        else:
            body.append(
                f'Content-Disposition: form-data; name="{fieldname}"\r\n\r\n'.encode()
            )

            body.append(value.encode("utf-8") if isinstance(value, str) else value)

        body.append(b"\r\n")

    body.append(f"--{boundary}--\r\n".encode())

    content_type = f"multipart/form-data; boundary={boundary}"

    return b"".join(body), content_type


@pytest.fixture
def http_client(http_client_class: type[HttpClient]) -> HttpClient:
    with contextlib.suppress(ImportError):
        from .clients.chalice import ChaliceHttpClient

        if http_client_class is ChaliceHttpClient:
            pytest.xfail(reason="Chalice does not support uploads")

    return http_client_class(schema)


@pytest.fixture
def enabled_http_client(http_client_class: type[HttpClient]) -> HttpClient:
    with contextlib.suppress(ImportError):
        from .clients.chalice import ChaliceHttpClient

        if http_client_class is ChaliceHttpClient:
            pytest.xfail(reason="Chalice does not support uploads")

    return http_client_class(schema, multipart_uploads_enabled=True)


async def test_multipart_uploads_are_disabled_by_default(http_client: HttpClient):
    f = BytesIO(b"strawberry")

    query = """
    mutation($textFile: Upload!) {
        readText(textFile: $textFile)
    }
    """

    response = await http_client.query(
        query,
        variables={"textFile": None},
        files={"textFile": f},
    )

    assert response.status_code == 400
    assert response.data == b"Unsupported content type"


async def test_upload(enabled_http_client: HttpClient):
    f = BytesIO(b"strawberry")

    query = """
    mutation($textFile: Upload!) {
        readText(textFile: $textFile)
    }
    """

    response = await enabled_http_client.query(
        query,
        variables={"textFile": None},
        files={"textFile": f},
    )

    assert response.json.get("errors") is None
    assert response.json["data"] == {"readText": "strawberry"}


async def test_file_list_upload(enabled_http_client: HttpClient):
    query = "mutation($files: [Upload!]!) { readFiles(files: $files) }"
    file1 = BytesIO(b"strawberry1")
    file2 = BytesIO(b"strawberry2")

    response = await enabled_http_client.query(
        query=query,
        variables={"files": [None, None]},
        files={"file1": file1, "file2": file2},
    )

    data = response.json["data"]
    assert isinstance(data, dict)
    assert len(data["readFiles"]) == 2
    assert data["readFiles"][0] == "strawberry1"
    assert data["readFiles"][1] == "strawberry2"


async def test_nested_file_list(enabled_http_client: HttpClient):
    query = "mutation($folder: FolderInput!) { readFolder(folder: $folder) }"
    file1 = BytesIO(b"strawberry1")
    file2 = BytesIO(b"strawberry2")

    response = await enabled_http_client.query(
        query=query,
        variables={"folder": {"files": [None, None]}},
        files={"file1": file1, "file2": file2},
    )

    data = response.json["data"]
    assert isinstance(data, dict)
    assert len(data["readFolder"]) == 2
    assert data["readFolder"][0] == "strawberry1"
    assert data["readFolder"][1] == "strawberry2"


async def test_upload_single_and_list_file_together(enabled_http_client: HttpClient):
    query = """
        mutation($files: [Upload!]!, $textFile: Upload!) {
            readFiles(files: $files)
            readText(textFile: $textFile)
        }
    """
    file1 = BytesIO(b"strawberry1")
    file2 = BytesIO(b"strawberry2")
    file3 = BytesIO(b"strawberry3")

    response = await enabled_http_client.query(
        query=query,
        variables={"files": [None, None], "textFile": None},
        files={"file1": file1, "file2": file2, "textFile": file3},
    )

    data = response.json["data"]
    assert isinstance(data, dict)
    assert len(data["readFiles"]) == 2
    assert data["readFiles"][0] == "strawberry1"
    assert data["readFiles"][1] == "strawberry2"
    assert data["readText"] == "strawberry3"


async def test_upload_invalid_query(enabled_http_client: HttpClient):
    f = BytesIO(b"strawberry")

    query = """
    mutation($textFile: Upload!) {
        readT
    """

    response = await enabled_http_client.query(
        query,
        variables={"textFile": None},
        files={"textFile": f},
    )

    assert response.status_code == 200
    assert response.json["data"] is None
    assert response.json["errors"] == [
        {
            "locations": [{"column": 5, "line": 4}],
            "message": "Syntax Error: Expected Name, found <EOF>.",
        }
    ]


async def test_upload_missing_file(enabled_http_client: HttpClient):
    f = BytesIO(b"strawberry")

    query = """
    mutation($textFile: Upload!) {
        readText(textFile: $textFile)
    }
    """

    response = await enabled_http_client.query(
        query,
        variables={"textFile": None},
        # using the wrong name to simulate a missing file
        # this is to make it easier to run tests with our client
        files={"a": f},
    )

    assert response.status_code == 400
    assert "File(s) missing in form data" in response.text


class FakeWriter:
    def __init__(self):
        self.buffer = BytesIO()

    async def write(self, data: bytes):
        self.buffer.write(data)

    @property
    def value(self) -> bytes:
        return self.buffer.getvalue()


async def test_extra_form_data_fields_are_ignored(enabled_http_client: HttpClient):
    query = """mutation($textFile: Upload!) {
        readText(textFile: $textFile)
    }"""

    f = BytesIO(b"strawberry")
    operations = json.dumps({"query": query, "variables": {"textFile": None}})
    file_map = json.dumps({"textFile": ["variables.textFile"]})
    extra_field_data = json.dumps({})

    f = BytesIO(b"strawberry")
    fields = {
        "operations": operations,
        "map": file_map,
        "extra_field": extra_field_data,
        "textFile": ("textFile.txt", f.read(), "text/plain"),
    }

    data, header = _encode_multipart_formdata(fields)

    response = await enabled_http_client.post(
        url="/graphql",
        data=data,
        headers={
            "content-type": header,
            "content-length": f"{len(data)}",
        },
    )

    assert response.status_code == 200
    assert response.json["data"] == {"readText": "strawberry"}


async def test_sending_invalid_form_data(enabled_http_client: HttpClient):
    headers = {"content-type": "multipart/form-data; boundary=----fake"}
    response = await enabled_http_client.post("/graphql", headers=headers)

    assert response.status_code == 400
    # TODO: consolidate this, it seems only AIOHTTP returns the second error
    # due to validating the boundary
    assert (
        "No GraphQL query found in the request" in response.text
        or "Unable to parse the multipart body" in response.text
    )


@pytest.mark.aiohttp
async def test_sending_invalid_json_body(enabled_http_client: HttpClient):
    f = BytesIO(b"strawberry")
    operations = "}"
    file_map = json.dumps({"textFile": ["variables.textFile"]})

    fields = {
        "operations": operations,
        "map": file_map,
        "textFile": ("textFile.txt", f.read(), "text/plain"),
    }

    data, header = _encode_multipart_formdata(fields)

    response = await enabled_http_client.post(
        "/graphql",
        data=data,
        headers={
            "content-type": header,
            "content-length": f"{len(data)}",
        },
    )

    assert response.status_code == 400
    assert (
        "Unable to parse the multipart body" in response.text
        or "Unable to parse request body as JSON" in response.text
    )



import textwrap

import strawberry
from strawberry.types.base import StrawberryObjectDefinition
from strawberry.types.field import StrawberryField


def test_can_change_which_fields_are_exposed():
    @strawberry.type
    class User:
        name: str
        email: str = strawberry.field(metadata={"tags": ["internal"]})

    @strawberry.type
    class Query:
        user: User

    def public_field_filter(field: StrawberryField) -> bool:
        return "internal" not in field.metadata.get("tags", [])

    class PublicSchema(strawberry.Schema):
        def get_fields(
            self, type_definition: StrawberryObjectDefinition
        ) -> list[StrawberryField]:
            fields = super().get_fields(type_definition)
            return list(filter(public_field_filter, fields))

    schema = PublicSchema(query=Query)

    expected_schema = textwrap.dedent(
        """
        type Query {
          user: User!
        }

        type User {
          name: String!
        }
        """
    ).strip()

    assert schema.as_str() == expected_schema



from typing import Literal

import pytest
from graphql import GraphQLError
from pytest_mock import MockFixture

from tests.conftest import skip_if_gql_32

from .clients.base import HttpClient


@pytest.mark.parametrize("method", ["get", "post"])
async def test_graphql_query(method: Literal["get", "post"], http_client: HttpClient):
    response = await http_client.query(
        method=method,
        query="{ hello }",
    )
    assert response.status_code == 200

    data = response.json["data"]
    assert isinstance(data, dict)
    assert data["hello"] == "Hello world"


@pytest.mark.parametrize("method", ["get", "post"])
async def test_calls_handle_errors(
    method: Literal["get", "post"], http_client: HttpClient, mocker: MockFixture
):
    sync_mock = mocker.patch(
        "strawberry.http.sync_base_view.SyncBaseHTTPView._handle_errors"
    )
    async_mock = mocker.patch(
        "strawberry.http.async_base_view.AsyncBaseHTTPView._handle_errors"
    )

    response = await http_client.query(
        method=method,
        query="{ hey }",
    )
    data = response.json["data"]

    assert response.status_code == 200
    assert data is None

    assert response.json["errors"] == [
        {
            "message": "Cannot query field 'hey' on type 'Query'.",
            "locations": [{"line": 1, "column": 3}],
        }
    ]

    error = GraphQLError("Cannot query field 'hey' on type 'Query'.")

    response_data = {
        "data": None,
        "errors": [
            {
                "message": "Cannot query field 'hey' on type 'Query'.",
                "locations": [{"line": 1, "column": 3}],
            },
        ],
        "extensions": {"example": "example"},
    }

    call_args = async_mock.call_args[0] if async_mock.called else sync_mock.call_args[0]

    assert call_args[0][0].message == error.message
    assert call_args[1] == response_data


@pytest.mark.parametrize("method", ["get", "post"])
async def test_graphql_can_pass_variables(
    method: Literal["get", "post"], http_client: HttpClient
):
    response = await http_client.query(
        method=method,
        query="query hello($name: String!) { hello(name: $name) }",
        variables={"name": "Jake"},
    )
    assert response.status_code == 200

    data = response.json["data"]
    assert isinstance(data, dict)
    assert data["hello"] == "Hello Jake"


@pytest.mark.parametrize("extra_kwargs", [{"variables": None}, {}])
async def test_operation_variables_may_be_null_or_omitted(
    http_client: HttpClient, extra_kwargs
):
    response = await http_client.query(
        query="{ __typename }",
        **extra_kwargs,
    )
    data = response.json["data"]

    assert response.status_code == 200
    assert isinstance(data, dict)
    assert data["__typename"] == "Query"


@pytest.mark.parametrize(
    "not_an_object_or_null",
    ["string", 0, False, ["array"]],
)
async def test_requests_with_invalid_variables_parameter_are_rejected(
    http_client: HttpClient, not_an_object_or_null
):
    response = await http_client.query(
        query="{ __typename }",
        variables=not_an_object_or_null,
    )

    assert response.status_code == 400
    assert (
        response.data
        == b"The GraphQL operation's `variables` must be an object or null, if provided."
    )


@pytest.mark.parametrize("method", ["get", "post"])
async def test_root_value(method: Literal["get", "post"], http_client: HttpClient):
    response = await http_client.query(
        method=method,
        query="{ rootName }",
    )
    assert response.status_code == 200

    data = response.json["data"]
    assert isinstance(data, dict)
    assert data["rootName"] == "Query"


@pytest.mark.parametrize("method", ["get", "post"])
async def test_passing_invalid_query(
    method: Literal["get", "post"], http_client: HttpClient
):
    response = await http_client.query(
        method=method,
        query="{ h",
    )

    assert response.status_code == 200
    assert response.json["errors"] == [
        {
            "message": "Syntax Error: Expected Name, found <EOF>.",
            "locations": [{"line": 1, "column": 4}],
        }
    ]


@pytest.mark.parametrize("method", ["get", "post"])
async def test_returns_errors(method: Literal["get", "post"], http_client: HttpClient):
    response = await http_client.query(
        method=method,
        query="{ maya }",
    )

    assert response.status_code == 200
    assert response.json["errors"] == [
        {
            "message": "Cannot query field 'maya' on type 'Query'.",
            "locations": [{"line": 1, "column": 3}],
        }
    ]


@pytest.mark.parametrize("method", ["get", "post"])
async def test_returns_errors_and_data(
    method: Literal["get", "post"], http_client: HttpClient
):
    response = await http_client.query(
        method=method,
        query="{ hello, alwaysFail }",
    )

    assert response.status_code == 200
    data = response.json["data"]
    errors = response.json["errors"]

    assert errors == [
        {
            "locations": [{"column": 10, "line": 1}],
            "message": "You are not authorized",
            "path": ["alwaysFail"],
        }
    ]
    assert data == {"hello": "Hello world", "alwaysFail": None}


async def test_passing_invalid_json_post(http_client: HttpClient):
    response = await http_client.post(
        url="/graphql",
        data=b"{ h",
        headers={"Content-Type": "application/json"},
    )

    assert response.status_code == 400
    assert "Unable to parse request body as JSON" in response.text


async def test_passing_invalid_json_get(http_client: HttpClient):
    response = await http_client.get(
        url="/graphql?query={ hello }&variables='{'",
    )

    assert response.status_code == 400
    assert "Unable to parse request body as JSON" in response.text


async def test_query_parameters_are_never_interpreted_as_list(http_client: HttpClient):
    response = await http_client.get(
        url='/graphql?query=query($name: String!) { hello(name: $name) }&variables={"name": "Jake"}&variables={"name": "Jake"}',
    )

    assert response.status_code == 200
    assert response.json["data"] == {"hello": "Hello Jake"}


async def test_missing_query(http_client: HttpClient):
    response = await http_client.post(
        url="/graphql",
        json={},
        headers={"Content-Type": "application/json"},
    )

    assert response.status_code == 400
    assert "No GraphQL query found in the request" in response.text


@pytest.mark.parametrize(
    "not_stringified_json",
    [{"obj": "ect"}, 0, False, ["array"]],
)
async def test_requests_with_invalid_query_parameter_are_rejected(
    http_client: HttpClient, not_stringified_json
):
    response = await http_client.query(
        query=not_stringified_json,
    )

    assert response.status_code == 400
    assert (
        response.data
        == b"The GraphQL operation's `query` must be a string or null, if provided."
    )


@pytest.mark.parametrize("method", ["get", "post"])
async def test_query_context(method: Literal["get", "post"], http_client: HttpClient):
    response = await http_client.query(
        method=method,
        query="{ valueFromContext }",
    )
    assert response.status_code == 200

    data = response.json["data"]
    assert isinstance(data, dict)
    assert data["valueFromContext"] == "a value from context"


@skip_if_gql_32("formatting is different in gql 3.2")
@pytest.mark.parametrize("method", ["get", "post"])
async def test_query_extensions(
    method: Literal["get", "post"], http_client: HttpClient
):
    response = await http_client.query(
        method=method,
        query='{ valueFromExtensions(key:"test") }',
        extensions={"test": "hello"},
    )
    assert response.status_code == 200

    data = response.json["data"]
    assert isinstance(data, dict)
    assert data["valueFromExtensions"] == "hello"


@pytest.mark.parametrize(
    "not_an_object_or_null",
    ["string", 0, False, ["array"]],
)
async def test_requests_with_invalid_extension_parameter_are_rejected(
    http_client: HttpClient, not_an_object_or_null
):
    response = await http_client.query(
        query="{ __typename }",
        extensions=not_an_object_or_null,
    )

    assert response.status_code == 400
    assert (
        response.data
        == b"The GraphQL operation's `extensions` must be an object or null, if provided."
    )


@pytest.mark.parametrize("method", ["get", "post"])
async def test_returning_status_code(
    method: Literal["get", "post"], http_client: HttpClient
):
    response = await http_client.query(
        method=method,
        query="{ returns401 }",
    )

    assert response.status_code == 401
    assert response.json["data"] == {"returns401": "hey"}


@pytest.mark.parametrize("method", ["get", "post"])
async def test_updating_headers(
    method: Literal["get", "post"], http_client: HttpClient
):
    response = await http_client.query(
        method=method,
        variables={"name": "Jake"},
        query="query ($name: String!) { setHeader(name: $name) }",
    )

    assert response.status_code == 200
    assert response.json["data"] == {"setHeader": "Jake"}
    assert response.headers["x-name"] == "Jake"


@pytest.mark.parametrize(
    ("extra_kwargs", "expected_message"),
    [
        ({}, "Hello Foo"),
        ({"operation_name": None}, "Hello Foo"),
        ({"operation_name": "Query1"}, "Hello Foo"),
        ({"operation_name": "Query2"}, "Hello Bar"),
    ],
)
async def test_operation_selection(
    http_client: HttpClient, extra_kwargs, expected_message
):
    response = await http_client.query(
        query="""
            query Query1 { hello(name: "Foo") }
            query Query2 { hello(name: "Bar") }
        """,
        **extra_kwargs,
    )

    assert response.status_code == 200
    assert response.json["data"] == {"hello": expected_message}


@pytest.mark.parametrize(
    "operation_name",
    ["", "Query3"],
)
async def test_invalid_operation_selection(http_client: HttpClient, operation_name):
    response = await http_client.query(
        query="""
            query Query1 { hello(name: "Foo") }
            query Query2 { hello(name: "Bar") }
        """,
        operation_name=operation_name,
    )

    assert response.status_code == 400
    assert response.data == f'Unknown operation named "{operation_name}".'.encode()


async def test_operation_selection_without_operations(http_client: HttpClient):
    response = await http_client.query(
        query="""
            fragment Fragment1 on Query { __typename }
        """,
    )

    assert response.status_code == 400
    assert response.data == b"Can't get GraphQL operation type"



import textwrap
from typing import ForwardRef

import pytest

import strawberry
from strawberry.parent import resolve_parent_forward_arg


def test_parent_type():
    global User

    try:

        def get_full_name(user: strawberry.Parent["User"]) -> str:
            return f"{user.first_name} {user.last_name}"

        @strawberry.type
        class User:
            first_name: str
            last_name: str
            full_name: str = strawberry.field(resolver=get_full_name)

        @strawberry.type
        class Query:
            @strawberry.field
            def user(self) -> User:
                return User(first_name="John", last_name="Doe")  # noqa: F821

        schema = strawberry.Schema(query=Query)

        expected = """\
            type Query {
              user: User!
            }

            type User {
              firstName: String!
              lastName: String!
              fullName: String!
            }
        """
        assert textwrap.dedent(str(schema)).strip() == textwrap.dedent(expected).strip()

        query = "{ user { firstName, lastName, fullName } }"
        result = schema.execute_sync(query)
        assert not result.errors
        assert result.data == {
            "user": {
                "firstName": "John",
                "lastName": "Doe",
                "fullName": "John Doe",
            }
        }
    finally:
        del User


@pytest.mark.parametrize(
    ("annotation", "expected"),
    [
        ("strawberry.Parent[str]", strawberry.Parent["str"]),
        ("Parent[str]", strawberry.Parent["str"]),
        (ForwardRef("strawberry.Parent[str]"), strawberry.Parent["str"]),
        (ForwardRef("Parent[str]"), strawberry.Parent["str"]),
        (strawberry.Parent["User"], strawberry.Parent["User"]),
    ],
)
def test_resolve_parent_forward_arg(annotation, expected):
    assert resolve_parent_forward_arg(annotation) == expected



from dataclasses import dataclass
from typing import Any

import pytest
from pytest_mock import MockerFixture

import strawberry
from strawberry.types.base import StrawberryObjectDefinition


def test_query_interface():
    @strawberry.interface
    class Cheese:
        name: str

    @strawberry.type
    class Swiss(Cheese):
        canton: str

    @strawberry.type
    class Italian(Cheese):
        province: str

    @strawberry.type
    class Root:
        @strawberry.field
        def assortment(self) -> list[Cheese]:
            return [
                Italian(name="Asiago", province="Friuli"),
                Swiss(name="Tomme", canton="Vaud"),
            ]

    schema = strawberry.Schema(query=Root, types=[Swiss, Italian])

    query = """{
        assortment {
            name
            ... on Italian { province }
            ... on Swiss { canton }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data is not None
    assert result.data["assortment"] == [
        {"name": "Asiago", "province": "Friuli"},
        {"canton": "Vaud", "name": "Tomme"},
    ]


def test_interfaces_can_implement_other_interfaces():
    @strawberry.interface
    class Error:
        message: str

    @strawberry.interface
    class FieldError(Error):
        message: str
        field: str

    @strawberry.type
    class PasswordTooShort(FieldError):
        message: str
        field: str
        fix: str

    @strawberry.type
    class Query:
        @strawberry.field
        def always_error(self) -> Error:
            return PasswordTooShort(
                message="Password Too Short",
                field="Password",
                fix="Choose more characters",
            )

    schema = strawberry.Schema(Query, types=[PasswordTooShort])
    query = """{
        alwaysError {
            ... on Error {
                message
            }
            ... on FieldError {
                field
            }
            ... on PasswordTooShort {
                fix
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data is not None
    assert result.data["alwaysError"] == {
        "message": "Password Too Short",
        "field": "Password",
        "fix": "Choose more characters",
    }


def test_interface_duck_typing():
    @strawberry.interface
    class Entity:
        id: int

    @strawberry.type
    class Anime(Entity):
        name: str

        @classmethod
        def is_type_of(cls, obj: Any, _) -> bool:
            return isinstance(obj, AnimeORM)

    @dataclass
    class AnimeORM:
        id: int
        name: str

    @strawberry.type
    class Query:
        @strawberry.field
        def anime(self) -> Entity:
            return AnimeORM(id=1, name="One Piece")  # type: ignore

    schema = strawberry.Schema(query=Query, types=[Anime])

    query = """{
        anime { id ... on Anime { name } }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"anime": {"id": 1, "name": "One Piece"}}


def test_interface_explicit_type_resolution():
    @dataclass
    class AnimeORM:
        id: int
        name: str

    @strawberry.interface
    class Node:
        id: int

    @strawberry.type
    class Anime(Node):
        name: str

        @classmethod
        def is_type_of(cls, obj: Any, _) -> bool:
            return isinstance(obj, AnimeORM)

    @strawberry.type
    class Query:
        @strawberry.field
        def node(self) -> Node:
            return AnimeORM(id=1, name="One Piece")  # type: ignore

    schema = strawberry.Schema(query=Query, types=[Anime])

    query = "{ node { __typename, id ... on Anime { name }} }"
    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {
        "node": {
            "__typename": "Anime",
            "id": 1,
            "name": "One Piece",
        }
    }


@pytest.mark.xfail(reason="We don't support returning dictionaries yet")
def test_interface_duck_typing_returning_dict():
    @strawberry.interface
    class Entity:
        id: int

    @strawberry.type
    class Anime(Entity):
        name: str

    @strawberry.type
    class Query:
        @strawberry.field
        def anime(self) -> Anime:
            return {"id": 1, "name": "One Piece"}  # type: ignore

    schema = strawberry.Schema(query=Query)

    query = """{
        anime { name }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"anime": {"name": "One Piece"}}


def test_duplicated_interface_in_multi_inheritance():
    """Test that interfaces are gathered properly via CPython's MRO.

    Previously interfaces were duplicated within a "Diamond Problem" inheritance
    scenario which is tested here. Using the MRO instead of the `__bases__` attribute of
    a class in :py:func:`strawberry.object_type._get_interfaces` allows Python's C3
    linearization algorithm to create a consistent precedents graph without duplicates.
    """

    @strawberry.interface
    class Base:
        id: str

    @strawberry.interface
    class InterfaceA(Base):
        id: str
        field_a: str

    @strawberry.interface
    class InterfaceB(Base):
        id: str
        field_b: str

    @strawberry.type
    class MyType(InterfaceA, InterfaceB):
        id: str
        field_a: str
        field_b: str

    @strawberry.type
    class Query:
        my_type: MyType

    type_definition: StrawberryObjectDefinition = MyType.__strawberry_definition__
    origins = [i.origin for i in type_definition.interfaces]
    assert origins == [InterfaceA, InterfaceB, Base]

    strawberry.Schema(Query)  # Final sanity check to ensure schema compiles


def test_interface_resolve_type(mocker: MockerFixture):
    """Check that the default implemenetation of `resolve_type` functions as expected.

    In this test-case the default implementation of `resolve_type` defined in
    `GraphQLCoreConverter.from_interface`, should immediately resolve the type of the
    returned concrete object. A concrete object is defined as one that is an instance of
    the interface it implements.

    Before the default implementation of `resolve_type`, the `is_type_of` methods of all
    specializations of an interface (in this case Anime & Movie) would be called. As
    this needlessly reduces performance, this test checks if only `Anime.is_type_of` is
    called when `Query.node` returns an `Anime` object.
    """

    class IsTypeOfTester:
        @classmethod
        def is_type_of(cls, obj: Any, _) -> bool:
            return isinstance(obj, cls)

    spy_is_type_of = mocker.spy(IsTypeOfTester, "is_type_of")

    @strawberry.interface
    class Node:
        id: int

    @strawberry.type
    class Anime(Node, IsTypeOfTester):
        name: str

    @strawberry.type
    class Movie(Node):
        title: str

        @classmethod
        def is_type_of(cls, *args: Any, **kwargs: Any) -> bool:
            del args, kwargs
            raise RuntimeError("Movie.is_type_of shouldn't have been called")

    @strawberry.type
    class Query:
        @strawberry.field
        def node(self) -> Node:
            return Anime(id=1, name="One Pierce")

    schema = strawberry.Schema(query=Query, types=[Anime, Movie])

    query = "{ node {  __typename, id } }"
    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"node": {"__typename": "Anime", "id": 1}}
    spy_is_type_of.assert_called_once()


def test_interface_specialized_resolve_type(mocker: MockerFixture):
    """Test that a specialized ``resolve_type`` is called."""

    class InterfaceTester:
        @classmethod
        def resolve_type(cls, obj: Any, *args: Any, **kwargs: Any) -> str:
            del args, kwargs
            return obj.__strawberry_definition__.name

    spy_resolve_type = mocker.spy(InterfaceTester, "resolve_type")

    @strawberry.interface
    class Food(InterfaceTester):
        id: int

    @strawberry.type
    class Fruit(Food):
        name: str

    @strawberry.type
    class Query:
        @strawberry.field
        def food(self) -> Food:
            return Fruit(id=1, name="strawberry")

    schema = strawberry.Schema(query=Query, types=[Fruit])
    result = schema.execute_sync("query { food { ... on Fruit { name } } }")

    assert not result.errors
    assert result.data == {"food": {"name": "strawberry"}}
    spy_resolve_type.assert_called_once()


@pytest.mark.asyncio
async def test_derived_interface(mocker: MockerFixture):
    """Test if correct resolve_type is called on a derived interface."""

    class NodeInterfaceTester:
        @classmethod
        def resolve_type(cls, obj: Any, *args: Any, **kwargs: Any) -> str:
            del args, kwargs
            return obj.__strawberry_definition__.name

    class NamedNodeInterfaceTester:
        @classmethod
        def resolve_type(cls, obj: Any, *args: Any, **kwargs: Any) -> str:
            del args, kwargs
            return obj.__strawberry_definition__.name

    spy_node_resolve_type = mocker.spy(NodeInterfaceTester, "resolve_type")
    spy_named_node_resolve_type = mocker.spy(NamedNodeInterfaceTester, "resolve_type")

    @strawberry.interface
    class Node(NodeInterfaceTester):
        id: int

    @strawberry.interface
    class NamedNode(NamedNodeInterfaceTester, Node):
        name: str

    @strawberry.type
    class Person(NamedNode):
        pass

    @strawberry.type
    class Query:
        @strawberry.field
        def friends(self) -> list[NamedNode]:
            return [Person(id=1, name="foo"), Person(id=2, name="bar")]

    schema = strawberry.Schema(Query, types=[Person])
    result = await schema.execute("query { friends { name } }")

    assert not result.errors
    assert result.data == {"friends": [{"name": "foo"}, {"name": "bar"}]}

    assert result.data is not None
    assert spy_named_node_resolve_type.call_count == len(result.data["friends"])
    spy_node_resolve_type.assert_not_called()


def test_resolve_type_on_interface_returning_interface():
    @strawberry.interface
    class Node:
        id: strawberry.ID

        @classmethod
        def resolve_type(cls, obj: Any, *args: Any, **kwargs: Any) -> str:
            return "Video" if obj.id == "1" else "Image"

    @strawberry.type
    class Video(Node): ...

    @strawberry.type
    class Image(Node): ...

    @strawberry.type
    class Query:
        @strawberry.field
        def node(self, id: strawberry.ID) -> Node:
            return Node(id=id)

    schema = strawberry.Schema(query=Query, types=[Video, Image])

    query = """
        query {
            one: node(id: "1") {
                __typename
                id
            }
            two: node(id: "2") {
                __typename
                id
            }
        }
    """

    result = schema.execute_sync(query)

    assert not result.errors

    assert result.data
    assert result.data["one"] == {"id": "1", "__typename": "Video"}
    assert result.data["two"] == {"id": "2", "__typename": "Image"}



from __future__ import annotations

from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING

import pytest

from tests.views.schema import schema

if TYPE_CHECKING:
    from strawberry.test import BaseGraphQLTestClient


@asynccontextmanager
async def aiohttp_graphql_client() -> AsyncGenerator[BaseGraphQLTestClient]:
    try:
        from aiohttp import web
        from aiohttp.test_utils import TestClient, TestServer

        from strawberry.aiohttp.test import GraphQLTestClient
        from strawberry.aiohttp.views import GraphQLView
    except ImportError:
        pytest.skip("Aiohttp not installed")

    view = GraphQLView(schema=schema)
    app = web.Application()
    app.router.add_route("*", "/graphql/", view)

    async with TestClient(TestServer(app)) as client:
        yield GraphQLTestClient(client)


@asynccontextmanager
async def asgi_graphql_client() -> AsyncGenerator[BaseGraphQLTestClient]:
    try:
        from starlette.testclient import TestClient

        from strawberry.asgi import GraphQL
        from strawberry.asgi.test import GraphQLTestClient
    except ImportError:
        pytest.skip("Starlette not installed")

    yield GraphQLTestClient(TestClient(GraphQL(schema)))


@asynccontextmanager
async def django_graphql_client() -> AsyncGenerator[BaseGraphQLTestClient]:
    try:
        from django.test.client import Client

        from strawberry.django.test import GraphQLTestClient
    except ImportError:
        pytest.skip("Django not installed")

    yield GraphQLTestClient(Client())


@pytest.fixture(
    params=[
        pytest.param(aiohttp_graphql_client, marks=[pytest.mark.aiohttp]),
        pytest.param(asgi_graphql_client, marks=[pytest.mark.asgi]),
        pytest.param(django_graphql_client, marks=[pytest.mark.django]),
    ]
)
async def graphql_client(request) -> AsyncGenerator[BaseGraphQLTestClient]:
    async with request.param() as graphql_client:
        yield graphql_client



from __future__ import annotations

from enum import Enum
from typing import Annotated

import strawberry


@strawberry.enum
class Role(Enum):
    ADMIN = "ADMIN"
    USER = "USER"
    GUEST = "GUEST"


@strawberry.interface
class Node:
    id: strawberry.ID


@strawberry.type
class PageInfo:
    has_next_page: bool
    has_previous_page: bool
    start_cursor: str | None
    end_cursor: str | None


@strawberry.type
class User(Node):
    id: strawberry.ID
    username: str
    email: str
    role: Role

    @classmethod
    def random(cls, seed: int) -> User:
        return User(
            id=strawberry.ID(str(int)),
            username=f"username={seed}",
            email=f"email={seed}",
            role=Role.ADMIN,
        )

    @strawberry.field
    def posts(self, first: int = 10, after: str | None = None) -> PostConnection | None:
        return PostConnection(
            edges=[
                PostEdge(
                    node=Post.random(i),
                    cursor=str(i),
                )
                for i in range(first)
            ],
            page_info=PageInfo(
                has_next_page=False,
                has_previous_page=False,
                start_cursor=None,
                end_cursor=None,
            ),
        )


@strawberry.type
class Post(Node):
    id: strawberry.ID
    title: str
    content: str
    author: User

    @classmethod
    def random(cls, seed: int) -> Post:
        return Post(
            id=strawberry.ID(str(int)),
            title=f"title={seed}",
            content=f"content={seed}",
            author=User.random(seed),
        )

    @strawberry.field
    def comments(
        self, first: int = 10, after: str | None = None
    ) -> CommentConnection | None:
        return CommentConnection(
            edges=[
                CommentEdge(
                    node=Comment.random(i),
                    cursor=str(i),
                )
                for i in range(first)
            ],
            page_info=PageInfo(
                has_next_page=False,
                has_previous_page=False,
                start_cursor=None,
                end_cursor=None,
            ),
        )


@strawberry.type
class Comment(Node):
    id: strawberry.ID
    text: str
    author: User
    post: Post

    @classmethod
    def random(cls, seed: int) -> Comment:
        return Comment(
            id=strawberry.ID(str(int)),
            text=f"text={seed}",
            author=User.random(seed),
            post=Post.random(seed),
        )


@strawberry.type
class UserConnection:
    edges: list[UserEdge | None] | None
    page_info: PageInfo


@strawberry.type
class UserEdge:
    node: User | None
    cursor: str


@strawberry.type
class PostConnection:
    edges: list[PostEdge | None] | None
    page_info: PageInfo


@strawberry.type
class PostEdge:
    node: Post | None
    cursor: str


@strawberry.type
class CommentConnection:
    edges: list[CommentEdge | None] | None
    page_info: PageInfo


@strawberry.type
class CommentEdge:
    node: Comment | None
    cursor: str


SearchResult = Annotated[User | Post | Comment, strawberry.union(name="SearchResult")]


@strawberry.type
class Query:
    users: UserConnection | None
    posts: PostConnection | None
    comments: CommentConnection | None

    @strawberry.field
    async def search(
        self, query: str, first: int = 10, after: str | None = None
    ) -> list[SearchResult | None] | None:
        div = 3

        chunks = [first // div + (1 if x < first % div else 0) for x in range(div)]

        return [
            *[User.random(i) for i in range(chunks[0])],
            *[Post.random(i) for i in range(chunks[1])],
            *[Comment.random(i) for i in range(chunks[2])],
        ]


schema = strawberry.Schema(query=Query)



import pytest

from strawberry.exceptions import MissingOptionalDependenciesError


def test_missing_optional_dependencies_error():
    with pytest.raises(MissingOptionalDependenciesError) as exc_info:
        raise MissingOptionalDependenciesError

    assert exc_info.value.message == "Some optional dependencies are missing"


def test_missing_optional_dependencies_error_packages():
    with pytest.raises(MissingOptionalDependenciesError) as exc_info:
        raise MissingOptionalDependenciesError(packages=["a", "b"])

    assert (
        exc_info.value.message
        == "Some optional dependencies are missing (hint: pip install a b)"
    )


def test_missing_optional_dependencies_error_empty_packages():
    with pytest.raises(MissingOptionalDependenciesError) as exc_info:
        raise MissingOptionalDependenciesError(packages=[])

    assert exc_info.value.message == "Some optional dependencies are missing"


def test_missing_optional_dependencies_error_extras():
    with pytest.raises(MissingOptionalDependenciesError) as exc_info:
        raise MissingOptionalDependenciesError(extras=["dev", "test"])

    assert (
        exc_info.value.message
        == "Some optional dependencies are missing (hint: pip install 'strawberry-graphql[dev,test]')"
    )


def test_missing_optional_dependencies_error_empty_extras():
    with pytest.raises(MissingOptionalDependenciesError) as exc_info:
        raise MissingOptionalDependenciesError(extras=[])

    assert exc_info.value.message == "Some optional dependencies are missing"


def test_missing_optional_dependencies_error_packages_and_extras():
    with pytest.raises(MissingOptionalDependenciesError) as exc_info:
        raise MissingOptionalDependenciesError(
            packages=["a", "b"],
            extras=["dev", "test"],
        )

    assert (
        exc_info.value.message
        == "Some optional dependencies are missing (hint: pip install a b 'strawberry-graphql[dev,test]')"
    )



from typing import Optional, Union

import strawberry
from strawberry.annotation import StrawberryAnnotation
from strawberry.types.base import StrawberryOptional
from strawberry.types.unset import UnsetType


def test_basic_optional():
    annotation = StrawberryAnnotation(Optional[str])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryOptional)
    assert resolved.of_type is str

    assert resolved == StrawberryOptional(of_type=str)
    assert resolved == Optional[str]


def test_optional_with_unset():
    annotation = StrawberryAnnotation(Union[UnsetType, str | None])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryOptional)
    assert resolved.of_type is str

    assert resolved == StrawberryOptional(of_type=str)
    assert resolved == Optional[str]


def test_optional_with_type_of_unset():
    annotation = StrawberryAnnotation(Union[type[strawberry.UNSET], str | None])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryOptional)
    assert resolved.of_type is str

    assert resolved == StrawberryOptional(of_type=str)
    assert resolved == Optional[str]


def test_optional_with_unset_as_union():
    annotation = StrawberryAnnotation(Union[UnsetType, None, str])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryOptional)
    assert resolved.of_type is str

    assert resolved == StrawberryOptional(of_type=str)
    assert resolved == Optional[str]


def test_optional_list():
    annotation = StrawberryAnnotation(Optional[list[bool]])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryOptional)
    assert resolved.of_type == list[bool]

    assert resolved == StrawberryOptional(of_type=list[bool])
    assert resolved == Optional[list[bool]]


def test_optional_optional():
    """Optional[Optional[...]] is squashed by Python to just Optional[...]"""
    annotation = StrawberryAnnotation(None | bool)
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryOptional)
    assert resolved.of_type is bool

    assert resolved == StrawberryOptional(of_type=bool)
    assert resolved == None | bool
    assert resolved == Optional[bool]


def test_optional_union():
    @strawberry.type
    class CoolType:
        foo: float

    @strawberry.type
    class UncoolType:
        bar: bool

    annotation = StrawberryAnnotation(Optional[CoolType | UncoolType])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryOptional)
    assert resolved.of_type == Union[CoolType, UncoolType]

    assert resolved == StrawberryOptional(of_type=Union[CoolType, UncoolType])
    assert resolved == Optional[CoolType | UncoolType]


# TODO: move to a field test file
def test_type_add_type_definition_with_fields():
    @strawberry.type
    class Query:
        name: str | None
        age: int | None

    definition = Query.__strawberry_definition__
    assert definition.name == "Query"

    [field1, field2] = definition.fields

    assert field1.python_name == "name"
    assert field1.graphql_name is None
    assert isinstance(field1.type, StrawberryOptional)
    assert field1.type.of_type is str

    assert field2.python_name == "age"
    assert field2.graphql_name is None
    assert isinstance(field2.type, StrawberryOptional)
    assert field2.type.of_type is int


# TODO: move to a field test file
def test_passing_custom_names_to_fields():
    @strawberry.type
    class Query:
        x: str | None = strawberry.field(name="name")
        y: int | None = strawberry.field(name="age")

    definition = Query.__strawberry_definition__
    assert definition.name == "Query"

    [field1, field2] = definition.fields

    assert field1.python_name == "x"
    assert field1.graphql_name == "name"
    assert isinstance(field1.type, StrawberryOptional)
    assert field1.type.of_type is str

    assert field2.python_name == "y"
    assert field2.graphql_name == "age"
    assert isinstance(field2.type, StrawberryOptional)
    assert field2.type.of_type is int


# TODO: move to a field test file
def test_passing_nothing_to_fields():
    @strawberry.type
    class Query:
        name: str | None = strawberry.field()
        age: int | None = strawberry.field()

    definition = Query.__strawberry_definition__
    assert definition.name == "Query"

    [field1, field2] = definition.fields

    assert field1.python_name == "name"
    assert field1.graphql_name is None
    assert isinstance(field1.type, StrawberryOptional)
    assert field1.type.of_type is str

    assert field2.python_name == "age"
    assert field2.graphql_name is None
    assert isinstance(field2.type, StrawberryOptional)
    assert field2.type.of_type is int


# TODO: move to a resolver test file
def test_resolver_fields():
    @strawberry.type
    class Query:
        @strawberry.field
        def name(self) -> str | None:
            return "Name"

    definition = Query.__strawberry_definition__
    assert definition.name == "Query"

    [field] = definition.fields

    assert field.python_name == "name"
    assert field.graphql_name is None
    assert isinstance(field.type, StrawberryOptional)
    assert field.type.of_type is str


# TODO: move to a resolver test file
def test_resolver_fields_arguments():
    @strawberry.type
    class Query:
        @strawberry.field
        def name(self, argument: str | None) -> str | None:
            return "Name"

    definition = Query.__strawberry_definition__

    assert definition.name == "Query"

    [field] = definition.fields

    assert field.python_name == "name"
    assert field.graphql_name is None
    assert isinstance(field.type, StrawberryOptional)
    assert field.type.of_type is str

    [argument] = field.arguments

    assert argument.python_name == "argument"
    assert argument.graphql_name is None
    assert isinstance(argument.type, StrawberryOptional)
    assert argument.type.of_type is str



import textwrap
from textwrap import dedent
from typing import Annotated

import strawberry
from strawberry.types.unset import UNSET


def test_argument_descriptions():
    @strawberry.type
    class Query:
        @strawberry.field
        def hello(  # type: ignore
            name: Annotated[
                str, strawberry.argument(description="Your name")
            ] = "Patrick",
        ) -> str:
            return f"Hi {name}"

    schema = strawberry.Schema(query=Query)

    assert str(schema) == dedent(
        '''\
        type Query {
          hello(
            """Your name"""
            name: String! = "Patrick"
          ): String!
        }'''
    )


def test_argument_deprecation_reason():
    @strawberry.type
    class Query:
        @strawberry.field
        def hello(  # type: ignore
            name: Annotated[
                str, strawberry.argument(deprecation_reason="Your reason")
            ] = "Patrick",
        ) -> str:
            return f"Hi {name}"

    schema = strawberry.Schema(query=Query)

    assert str(schema) == dedent(
        """\
        type Query {
          hello(name: String! = "Patrick" @deprecated(reason: "Your reason")): String!
        }"""
    )


def test_argument_names():
    @strawberry.input
    class HelloInput:
        name: str = strawberry.field(default="Patrick", description="Your name")

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(
            self, input_: Annotated[HelloInput, strawberry.argument(name="input")]
        ) -> str:
            return f"Hi {input_.name}"

    schema = strawberry.Schema(query=Query)

    assert str(schema) == dedent(
        '''\
        input HelloInput {
          """Your name"""
          name: String! = "Patrick"
        }

        type Query {
          hello(input: HelloInput!): String!
        }'''
    )


def test_argument_with_default_value_none():
    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self, name: str | None = None) -> str:
            return f"Hi {name}"

    schema = strawberry.Schema(query=Query)

    assert str(schema) == dedent(
        """\
        type Query {
          hello(name: String = null): String!
        }"""
    )


def test_optional_argument_unset():
    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self, name: str | None = UNSET, age: int | None = UNSET) -> str:
            if name is UNSET:
                return "Hi there"
            return f"Hi {name}"

    schema = strawberry.Schema(query=Query)

    assert str(schema) == dedent(
        """\
        type Query {
          hello(name: String, age: Int): String!
        }"""
    )

    result = schema.execute_sync(
        """
        query {
            hello
        }
    """
    )
    assert not result.errors
    assert result.data == {"hello": "Hi there"}


def test_optional_input_field_unset():
    @strawberry.input
    class TestInput:
        name: str | None = UNSET
        age: int | None = UNSET

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self, input: TestInput) -> str:
            if input.name is UNSET:
                return "Hi there"
            return f"Hi {input.name}"

    schema = strawberry.Schema(query=Query)

    assert (
        str(schema)
        == dedent(
            """
        type Query {
          hello(input: TestInput!): String!
        }

        input TestInput {
          name: String
          age: Int
        }
        """
        ).strip()
    )

    result = schema.execute_sync(
        """
        query {
            hello(input: {})
        }
    """
    )
    assert not result.errors
    assert result.data == {"hello": "Hi there"}


def test_setting_metadata_on_argument():
    field_definition = None

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(
            self,
            info: strawberry.Info,
            input: Annotated[str, strawberry.argument(metadata={"test": "foo"})],
        ) -> str:
            nonlocal field_definition
            field_definition = info._field
            return f"Hi {input}"

    schema = strawberry.Schema(query=Query)

    result = schema.execute_sync(
        """
        query {
            hello(input: "there")
        }
    """
    )
    assert not result.errors
    assert result.data == {"hello": "Hi there"}

    assert field_definition
    assert field_definition.arguments[0].metadata == {
        "test": "foo",
    }


def test_argument_parse_order():
    """Check early early exit from argument parsing due to finding ``info``.

    Reserved argument parsing, which interally also resolves annotations, exits early
    after detecting the ``info`` argumnent. As a result, the annotation of the ``id_``
    argument in `tests.schema.test_annotated.type_a.Query` is never resolved. This
    results in `StrawberryArgument` not being able to detect that ``id_`` makes use of
    `typing.Annotated` and `strawberry.argument`.

    This behavior is fixed by by ensuring that `StrawberryArgument` makes use of the new
    `StrawberryAnnotation.evaluate` method instead of consuming the raw annotation.

    An added benefit of this fix is that by removing annotation resolving code from
    `StrawberryResolver` and making it a part of `StrawberryAnnotation`, it makes it
    possible for `StrawberryArgument` and `StrawberryResolver` to share the same type
    evaluation cache.

    Refer to: https://github.com/strawberry-graphql/strawberry/issues/2855
    """
    from tests.schema.test_annotated import type_a, type_b

    expected = """
    type Query {
      getTesting(id: UUID!): String
    }

    scalar UUID
    """

    schema_a = strawberry.Schema(type_a.Query)
    schema_b = strawberry.Schema(type_b.Query)

    assert str(schema_a) == str(schema_b)
    assert str(schema_a) == textwrap.dedent(expected).strip()



import json
from typing import Literal

import pytest

from strawberry.http.base import BaseView

from .clients.base import HttpClient


@pytest.mark.parametrize("method", ["delete", "head", "put", "patch"])
async def test_does_only_allow_get_and_post(
    method: Literal["delete", "head", "put", "patch"],
    http_client: HttpClient,
):
    response = await http_client.request(url="/graphql", method=method)

    assert response.status_code == 405


async def test_the_http_handler_uses_the_views_decode_json_method(
    http_client: HttpClient, mocker
):
    spy = mocker.spy(BaseView, "decode_json")

    response = await http_client.query(query="{ hello }")
    assert response.status_code == 200
    assert response.headers["content-type"].split(";")[0] == "application/json"

    data = response.json["data"]
    assert isinstance(data, dict)
    assert data["hello"] == "Hello world"

    assert spy.call_count == 1


async def test_the_http_handler_supports_bytes_encoded_json(
    http_client: HttpClient, mocker
):
    """Check that http handlers correctly deal with byte return type from `encode_json`"""

    def patched_encode_json(self, data: object) -> bytes:
        return json.dumps(data).encode()

    mocker.patch("strawberry.http.base.BaseView.encode_json", patched_encode_json)

    response = await http_client.query(query="{ hello }")
    assert response.status_code == 200
    assert response.headers["content-type"].split(";")[0] == "application/json"

    data = response.json["data"]
    assert isinstance(data, dict)
    assert data["hello"] == "Hello world"


async def test_exception(http_client: HttpClient, mocker):
    response = await http_client.query(query="{ hello }", operation_name="wrong")
    assert response.status_code == 400
    assert response.headers["content-type"].split(";")[0] == "text/plain"
    assert response.data == b'Unknown operation named "wrong".'



import re
from collections.abc import Callable
from typing import Annotated, Any

import pytest

import strawberry
from strawberry.extensions.field_extension import (
    AsyncExtensionResolver,
    FieldExtension,
    SyncExtensionResolver,
)
from strawberry.schema.config import StrawberryConfig


class UpperCaseExtension(FieldExtension):
    def resolve(
        self,
        next_: Callable[..., Any],
        source: Any,
        info: strawberry.Info,
        **kwargs: Any,
    ):
        result = next_(source, info, **kwargs)
        return str(result).upper()


class LowerCaseExtension(FieldExtension):
    def resolve(
        self,
        next_: Callable[..., Any],
        source: Any,
        info: strawberry.Info,
        **kwargs: Any,
    ):
        result = next_(source, info, **kwargs)
        return str(result).lower()


class AsyncUpperCaseExtension(FieldExtension):
    async def resolve_async(
        self,
        next_: AsyncExtensionResolver,
        source: Any,
        info: strawberry.Info,
        **kwargs: Any,
    ):
        result = await next_(source, info, **kwargs)
        return str(result).upper()


class IdentityExtension(FieldExtension):
    def resolve(
        self,
        next_: SyncExtensionResolver,
        source: Any,
        info: strawberry.Info,
        **kwargs: Any,
    ) -> Any:
        return next_(source, info, **kwargs)

    async def resolve_async(
        self,
        next_: AsyncExtensionResolver,
        source: Any,
        info: strawberry.Info,
        **kwargs: Any,
    ) -> Any:
        return await next_(source, info, **kwargs)


def test_extension_argument_modification():
    @strawberry.type
    class Query:
        @strawberry.field(extensions=[UpperCaseExtension()])
        def string(self) -> str:
            return "This is a test!!"

    schema = strawberry.Schema(query=Query)
    query = "query { string }"

    result = schema.execute_sync(query)
    # The result should be lowercase because that is the last extension in the chain
    assert result.data["string"] == "THIS IS A TEST!!"


def test_extension_result_modification_sync():
    @strawberry.type
    class Query:
        @strawberry.field(extensions=[UpperCaseExtension()])
        def string(self) -> str:
            return "This is a test!!"

    schema = strawberry.Schema(query=Query)
    query = "query { string }"

    result = schema.execute_sync(query)
    # The result should be lowercase because that is the last extension in the chain
    assert result.data["string"] == "THIS IS A TEST!!"


async def test_async_extension_on_sync_resolver():
    @strawberry.type
    class Query:
        @strawberry.field(extensions=[AsyncUpperCaseExtension()])
        def string(self) -> str:
            return "This is a test!!"

    schema = strawberry.Schema(query=Query)
    query = "query { string }"

    result = await schema.execute(query)
    # The result should be lowercase because that is the last extension in the chain
    assert result.data["string"] == "THIS IS A TEST!!"


async def test_extension_result_modification_async():
    @strawberry.type
    class Query:
        @strawberry.field(extensions=[AsyncUpperCaseExtension()])
        async def string(self) -> str:
            return "This is a test!!"

    schema = strawberry.Schema(query=Query)
    query = "query { string }"

    result = await schema.execute(query)

    # The result should be lowercase because that is the last extension in the chain
    assert result.data["string"] == "THIS IS A TEST!!"


def test_fail_cannot_use_async_before_sync_extensions():
    @strawberry.type
    class Query:
        @strawberry.field(extensions=[AsyncUpperCaseExtension(), LowerCaseExtension()])
        def string(self) -> str:
            return "This is a test!!"

    # LowerCaseExtension should work just fine because it's sync
    msg = (
        "Query fields cannot be resolved. Cannot mix async-only extension(s) "
        "AsyncUpperCaseExtension with sync-only extension(s) "
        "LowerCaseExtension on Field string. "
        "If possible try to change the execution order so that all sync-only "
        "extensions are executed first."
    )
    with pytest.raises(TypeError, match=re.escape(msg)):
        strawberry.Schema(query=Query)


async def test_can_use_sync_before_async_extensions():
    @strawberry.type
    class Query:
        @strawberry.field(extensions=[LowerCaseExtension(), AsyncUpperCaseExtension()])
        def string(self) -> str:
            return "This is a test!!"

    schema = strawberry.Schema(query=Query)
    query = "query { string }"

    result = await schema.execute(query)
    # The result should be lowercase because that is the last extension in the chain
    assert result.data["string"] == "THIS IS A TEST!!"


async def test_can_use_sync_only_and_sync_before_async_extensions():
    """Use Sync - Sync + Async - Sync - Async possible."""

    @strawberry.type
    class Query:
        @strawberry.field(
            extensions=[
                LowerCaseExtension(),
                IdentityExtension(),
                LowerCaseExtension(),
                AsyncUpperCaseExtension(),
            ]
        )
        def string(self) -> str:
            return "This is a test!!"

    schema = strawberry.Schema(query=Query)
    query = "query { string }"

    result = await schema.execute(query)
    # The result should be lowercase because that is the last extension in the chain
    assert result.data["string"] == "THIS IS A TEST!!"


def test_fail_on_missing_async_extensions():
    class LowerCaseExtension(FieldExtension):
        def resolve(
            self,
            next_: Callable[..., Any],
            source: Any,
            info: strawberry.Info,
            **kwargs: Any,
        ):
            result = next_(source, info, **kwargs)
            return str(result).lower()

    class UpperCaseExtension(FieldExtension):
        async def resolve_async(
            self,
            next_: Callable[..., Any],
            source: Any,
            info: strawberry.Info,
            **kwargs: Any,
        ):
            result = await next_(source, info, **kwargs)
            return str(result).upper()

    @strawberry.type
    class Query:
        @strawberry.field(extensions=[UpperCaseExtension(), LowerCaseExtension()])
        async def string(self) -> str:
            return "This is a test!!"

    # UpperCaseExtension should work just fine because it's sync
    msg = (
        "Query fields cannot be resolved. Cannot add sync-only extension(s) "
        "LowerCaseExtension to the async resolver of Field string. "
        "Please add a resolve_async method to the extension(s)."
    )
    with pytest.raises(TypeError, match=re.escape(msg)):
        strawberry.Schema(query=Query)


def test_extension_order_respected():
    class LowerCaseExtension(FieldExtension):
        def resolve(
            self,
            next_: Callable[..., Any],
            source: Any,
            info: strawberry.Info,
            **kwargs: Any,
        ):
            result = next_(source, info, **kwargs)
            return str(result).lower()

    class UpperCaseExtension(FieldExtension):
        def resolve(
            self,
            next_: Callable[..., Any],
            source: Any,
            info: strawberry.Info,
            **kwargs: Any,
        ):
            result = next_(source, info, **kwargs)
            return str(result).upper()

    @strawberry.type
    class Query:
        @strawberry.field(extensions=[UpperCaseExtension(), LowerCaseExtension()])
        def string(self) -> str:
            return "This is a test!!"

    schema = strawberry.Schema(query=Query)
    query = "query { string }"

    result = schema.execute_sync(query)
    # The result should be lowercase because that is the last extension in the chain
    assert result.data["string"] == "this is a test!!"


def test_extension_argument_parsing():
    """Check that kwargs passed to field extensions have been converted into
    Strawberry types.
    """

    @strawberry.input
    class StringInput:
        some_input_value: str = strawberry.field(description="foo")

    field_kwargs = {}

    class CustomExtension(FieldExtension):
        def resolve(
            self,
            next_: Callable[..., Any],
            source: Any,
            info: strawberry.Info,
            **kwargs: Any,
        ):
            nonlocal field_kwargs
            field_kwargs = kwargs
            return next_(source, info, **kwargs)

    @strawberry.type
    class Query:
        @strawberry.field(extensions=[CustomExtension()])
        def string(self, some_input: StringInput) -> str:
            return f"This is a test!! {some_input.some_input_value}"

    schema = strawberry.Schema(query=Query)
    query = 'query { string(someInput: { someInputValue: "foo" }) }'

    result = schema.execute_sync(query)
    assert result.data, result.errors
    assert result.data["string"] == "This is a test!! foo"

    assert isinstance(field_kwargs["some_input"], StringInput)
    input_value = field_kwargs["some_input"]
    assert input_value.some_input_value == "foo"
    assert input_value.__strawberry_definition__.is_input is True


def test_extension_mutate_arguments():
    class CustomExtension(FieldExtension):
        def resolve(
            self,
            next_: Callable[..., Any],
            source: Any,
            info: strawberry.Info,
            **kwargs: Any,
        ):
            kwargs["some_input"] += 10
            return next_(source, info, **kwargs)

    @strawberry.type
    class Query:
        @strawberry.field(extensions=[CustomExtension()])
        def string(self, some_input: int) -> str:
            return f"This is a test!! {some_input}"

    schema = strawberry.Schema(query=Query)
    query = "query { string(someInput: 3) }"

    result = schema.execute_sync(query)
    assert result.data, result.errors
    assert result.data["string"] == "This is a test!! 13"


def test_extension_access_argument_metadata():
    field_kwargs = {}
    argument_metadata = {}

    class CustomExtension(FieldExtension):
        def resolve(
            self,
            next_: Callable[..., Any],
            source: Any,
            info: strawberry.Info,
            **kwargs: Any,
        ):
            nonlocal field_kwargs
            field_kwargs = kwargs

            for key in kwargs:
                argument_def = info.get_argument_definition(key)
                assert argument_def is not None
                argument_metadata[key] = argument_def.metadata

            return next_(source, info, **kwargs)

    @strawberry.type
    class Query:
        @strawberry.field(extensions=[CustomExtension()])
        def string(
            self,
            some_input: Annotated[str, strawberry.argument(metadata={"test": "foo"})],
            another_input: str | None = None,
        ) -> str:
            return f"This is a test!! {some_input}"

    schema = strawberry.Schema(query=Query)
    query = 'query { string(someInput: "foo") }'

    result = schema.execute_sync(query)
    assert result.data, result.errors
    assert result.data["string"] == "This is a test!! foo"

    assert isinstance(field_kwargs["some_input"], str)
    assert argument_metadata == {
        "some_input": {
            "test": "foo",
        },
        "another_input": {},
    }


def test_extension_has_custom_info_class():
    class CustomInfo(strawberry.Info):
        test: str = "foo"

    class CustomExtension(FieldExtension):
        def resolve(
            self,
            next_: Callable[..., Any],
            source: Any,
            info: CustomInfo,
            **kwargs: Any,
        ):
            assert isinstance(info, CustomInfo)
            # Explicitly check it's not Info.
            assert strawberry.Info in type(info).__bases__
            assert info.test == "foo"
            return next_(source, info, **kwargs)

    @strawberry.type
    class Query:
        @strawberry.field(extensions=[CustomExtension()])
        def string(self) -> str:
            return "This is a test!!"

    schema = strawberry.Schema(
        query=Query, config=StrawberryConfig(info_class=CustomInfo)
    )
    query = "query { string }"
    result = schema.execute_sync(query)
    assert result.data, result.errors
    assert result.data["string"] == "This is a test!!"



import contextlib
from typing import Literal

import pytest

import strawberry
from strawberry.http.base import BaseView
from strawberry.schema.config import StrawberryConfig
from tests.http.clients.base import HttpClient
from tests.views.schema import Mutation, MyExtension, Query, Subscription, schema


@pytest.fixture
def http_client(http_client_class: type[HttpClient]) -> HttpClient:
    with contextlib.suppress(ImportError):
        import django

        if django.VERSION < (4, 2):
            pytest.skip(reason="Django < 4.2 doesn't async streaming responses")

        from tests.http.clients.django import DjangoHttpClient

        if http_client_class is DjangoHttpClient:
            pytest.skip(
                reason="(sync) DjangoHttpClient doesn't support multipart subscriptions"
            )

    with contextlib.suppress(ImportError):
        from tests.http.clients.channels import SyncChannelsHttpClient

        # TODO: why do we have a sync channels client?
        if http_client_class is SyncChannelsHttpClient:
            pytest.skip(
                reason="SyncChannelsHttpClient doesn't support multipart subscriptions"
            )

    with contextlib.suppress(ImportError):
        from tests.http.clients.async_flask import AsyncFlaskHttpClient
        from tests.http.clients.flask import FlaskHttpClient

        if http_client_class is FlaskHttpClient:
            pytest.skip(
                reason="FlaskHttpClient doesn't support multipart subscriptions"
            )

        if http_client_class is AsyncFlaskHttpClient:
            pytest.xfail(
                reason="AsyncFlaskHttpClient doesn't support multipart subscriptions"
            )

    with contextlib.suppress(ImportError):
        from tests.http.clients.chalice import ChaliceHttpClient

        if http_client_class is ChaliceHttpClient:
            pytest.skip(
                reason="ChaliceHttpClient doesn't support multipart subscriptions"
            )

    return http_client_class(schema=schema)


@pytest.mark.parametrize("method", ["get", "post"])
@pytest.mark.parametrize(
    "accept_header",
    [
        pytest.param(
            "multipart/mixed;boundary=graphql;subscriptionSpec=1.0,application/json",
            id="with-boundary",
        ),
        pytest.param(
            'multipart/mixed;subscriptionSpec="1.0",application/json',
            id="no-boundary-with-quotes",
        ),
    ],
)
async def test_multipart_subscription(
    http_client: HttpClient, method: Literal["get", "post"], accept_header: str
):
    response = await http_client.query(
        method=method,
        query='subscription { echo(message: "Hello world", delay: 0.2) }',
        headers={
            "accept": accept_header,
            "content-type": "application/json",
        },
    )

    data = [d async for d in response.streaming_json()]

    assert data == [
        {
            "payload": {
                "data": {"echo": "Hello world"},
                "extensions": {"example": "example"},
            }
        }
    ]

    assert response.status_code == 200


async def test_multipart_subscription_use_the_views_decode_json_method(
    http_client: HttpClient, mocker
):
    spy = mocker.spy(BaseView, "decode_json")

    response = await http_client.query(
        query='subscription { echo(message: "Hello world", delay: 0.2) }',
        headers={
            "accept": "multipart/mixed;boundary=graphql;subscriptionSpec=1.0,application/json",
            "content-type": "application/json",
        },
    )

    data = [d async for d in response.streaming_json()]

    assert data == [
        {
            "payload": {
                "data": {"echo": "Hello world"},
                "extensions": {"example": "example"},
            }
        }
    ]

    assert response.status_code == 200

    assert spy.call_count == 1


async def test_returns_error_when_trying_to_use_batching_with_multipart_subscriptions(
    http_client_class: type[HttpClient],
):
    http_client = http_client_class(
        schema=strawberry.Schema(
            query=Query,
            mutation=Mutation,
            subscription=Subscription,
            extensions=[MyExtension],
            config=StrawberryConfig(batching_config={"max_operations": 10}),
        )
    )

    response = await http_client.post(
        url="/graphql",
        json=[
            {"query": 'subscription { echo(message: "Hello world", delay: 0.2) }'},
            {"query": 'subscription { echo(message: "Hello world", delay: 0.2) }'},
            {"query": 'subscription { echo(message: "Hello world", delay: 0.2) }'},
        ],
        headers={
            "content-type": "application/json",
            "accept": "multipart/mixed;boundary=graphql;subscriptionSpec=1.0,application/json",
        },
    )

    assert response.status_code == 400
    assert "Batching is not supported for multipart subscriptions" in response.text



import importlib
import sys
import textwrap
import warnings

import pytest

import strawberry
from tests.experimental.pydantic.utils import needs_pydantic_v2


@pytest.mark.skipif(
    sys.version_info >= (3, 14),
    reason="Pydantic v1 is not compatible with Python 3.14+",
)
@needs_pydantic_v2
def test_can_use_both_pydantic_1_and_2():
    import pydantic
    from pydantic import v1 as pydantic_v1

    class UserModel(pydantic.BaseModel):
        age: int
        name: str | None

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age: strawberry.auto
        name: strawberry.auto

    class LegacyUserModel(pydantic_v1.BaseModel):
        age: int
        name: str | None
        int_field: pydantic.v1.NonNegativeInt = 1

    @strawberry.experimental.pydantic.type(LegacyUserModel)
    class LegacyUser:
        age: strawberry.auto
        name: strawberry.auto
        int_field: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self, id: strawberry.ID) -> User | LegacyUser:
            if id == "legacy":
                return LegacyUser(age=1, name="legacy")

            return User(age=1, name="ABC")

    schema = strawberry.Schema(query=Query)

    expected_schema = """
    type LegacyUser {
      age: Int!
      name: String
      intField: Int!
    }

    type Query {
      user(id: ID!): UserLegacyUser!
    }

    type User {
      age: Int!
      name: String
    }

    union UserLegacyUser = User | LegacyUser
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()

    query = """
        query ($id: ID!) {
            user(id: $id) {
                __typename
                ... on User { name }
                ... on LegacyUser { name }
            }
        }
    """

    result = schema.execute_sync(query, variable_values={"id": "new"})

    assert not result.errors
    assert result.data == {"user": {"__typename": "User", "name": "ABC"}}

    result = schema.execute_sync(query, variable_values={"id": "legacy"})

    assert not result.errors
    assert result.data == {"user": {"__typename": "LegacyUser", "name": "legacy"}}


@pytest.mark.skipif(
    sys.version_info >= (3, 14),
    reason="Pydantic v1 is not compatible with Python 3.14+",
)
@needs_pydantic_v2
def test_can_use_nested_pydantic_v1_models():
    from pydantic import v1 as pydantic_v1

    class Book(pydantic_v1.BaseModel):
        title: str

    class Library(pydantic_v1.BaseModel):
        books: list[Book]

    @strawberry.experimental.pydantic.type(model=Book, all_fields=True)
    class BookType:
        pass

    @strawberry.experimental.pydantic.type(model=Library, all_fields=True)
    class LibraryType:
        pass

    @strawberry.type
    class Query:
        library: LibraryType

    schema = strawberry.Schema(query=Query)

    expected_schema = """
    type BookType {
      title: String!
    }

    type LibraryType {
      books: [BookType!]!
    }

    type Query {
      library: LibraryType!
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()


@pytest.mark.skipif(
    sys.version_info < (3, 14),
    reason="This test is only relevant on Python 3.14+",
)
@needs_pydantic_v2
def test_no_pydantic_v1_warning_on_python_314():
    import strawberry.experimental.pydantic._compat as compat_module

    with warnings.catch_warnings(record=True) as caught:
        warnings.simplefilter("always")
        importlib.reload(compat_module)

    pydantic_v1_warnings = [w for w in caught if "Pydantic V1" in str(w.message)]
    assert pydantic_v1_warnings == []
